/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 *
 * BioHDF library functions. 
 *
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>

#include "hdf5.h"

#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

int _create_sequence_dataset(hid_t gid, int seq_length);

int _create_quality_dataset(hid_t gid, int length);

int _create_seqid_dataset(hid_t gid, int seqid_length);

void _write_hit_buf(hit_buf_t *hit_buf, int hit_buf_num);

/* Used by the NCList indexing code to determine containment */
int _is_contained(refmap_t first, refmap_t second);

static void
_nclist_sort(refmap_t *alignments, hsize_t n_alignments,
             refmap_t *index /*OUT*/,
             nclist_level_t **levels /*OUT*/, hsize_t *curr_level /*OUT*/);
             
static void
_nclist_sort_helper(refmap_t *alignments, hsize_t n_alignments,
                    refmap_t *index /*OUT*/,
                    nclist_level_t **levels /*OUT*/, hsize_t *curr_level /*OUT*/,
                    hsize_t *max_levels /*OUT*/, hsize_t *index_offset /*OUT*/);

/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

/* TODO - parse multi-line fasta properly */
int H5BIOparse_fastA(hid_t gid_seqs, int max_seq_length, int is_verbose) {

  int seq_length = -1; /* parsed from first sequence */
  hid_t did_seqid;
  hid_t did_seq; /* created once length known */

  /* create buffer for sequence and seqid data */
  int seq_buf_size = 1024;
  char* seq_buf = NULL;
  int seqid_buf_size = 1024;
  char* seqid_buf = (char*)calloc(seqid_buf_size * SEQID_LENGTH, sizeof(char));

  /* create the seqid dataset */
  did_seqid = _create_seqid_dataset(gid_seqs, SEQID_LENGTH);

  /* parse the csfasta file */
  char line[1024]; /* TODO - parameterize max line length */
  char *the_seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char *seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char seq[SEQ_MAX_LENGTH];
  int seq_num = 0;

  clock_t start_clock, end_clock;
  if (is_verbose) { start_clock = clock(); }

  strcpy(seqid, "");
  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip comments */
    if (line[0] == '#') {
      continue;
    }

    /* parse id */
    if (sscanf(line, ">%s", the_seqid) == 1) {

      /* append previous sequence to dataset */
      if (strlen(seqid) > 0) {

	/* append the seqid to dataset, via a buffer */
	int seqid_buf_loc = seq_num % seqid_buf_size;
	memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
	if (seqid_buf_loc == seqid_buf_size - 1) {
	  H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_size, SEQID_LENGTH);
	}

	/* append the sequence to dataset, via a buffer */
	int buf_loc = seq_num % seq_buf_size;
	memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
	if (buf_loc == seq_buf_size - 1) {
	  H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, seq_buf_size, seq_length);
	}

	++seq_num;

	if (is_verbose) {
	  if (seq_num % 1000000 == 0) {
	    end_clock = clock();
	    double elapsed = ((double)(end_clock - start_clock)) / CLOCKS_PER_SEC;
	    start_clock = end_clock;
	    fprintf(stderr, "Processed %d sequences in %g seconds\n", seq_num, elapsed);
	  }
	}

      }
      memset(seqid, 0, SEQID_LENGTH);
      strcpy(seqid, the_seqid);
    } 
    else {
      memset(seq, 0, SEQ_MAX_LENGTH);
      sscanf(line, "%s", seq);

      if (seq_length < 0) {
	if (max_seq_length < 0) {
	  seq_length = strlen(seq);
	} else {
	  seq_length = max_seq_length;
	}

	/* create the sequence dataset and write buffer */
        did_seq = _create_sequence_dataset(gid_seqs, seq_length);
	seq_buf = (char*)calloc(seq_buf_size * seq_length, sizeof(char));
      }

    }
  }

  /* append previous sequence to dataset */
  if (strlen(seqid) > 0) {
    
    /* append the seqid to dataset, via a buffer */
    int seqid_buf_loc = seq_num % seqid_buf_size;
    memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
    H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_loc + 1, SEQID_LENGTH);
      
    /* append the sequence to dataset, via a buffer */
    int buf_loc = seq_num % seq_buf_size;
    memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
    H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, buf_loc + 1, seq_length);
      
    ++seq_num;
  }
  if (did_seq) { H5Dclose(did_seq); }

  if (is_verbose) { fprintf(stderr, "Processed %d sequences\n", seq_num); }

  return 0;
}


int H5BIOparse_fastQ(hid_t gid_seqs, int max_seq_length, int is_verbose) {

  int seq_length = -1; /* parsed from first sequence */
  hid_t did_seqid;
  hid_t did_seq; /* created once length known */
  hid_t did_qual; /* created once length known */

  /* create buffer for sequence, quality and seqid data */
  int seq_buf_size = 1024;
  char* seq_buf = NULL;
  char* qual_buf = NULL;
  int seqid_buf_size = 1024;
  char* seqid_buf = (char*)calloc(seqid_buf_size * SEQID_LENGTH, sizeof(char));

  /* create the seqid dataset */
  did_seqid = _create_seqid_dataset(gid_seqs, SEQID_LENGTH);

  /* parse the fastq file */
  char line[1024]; /* TODO - parameterize max line length */
  char *seqid = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char *seqid2 = (char *)calloc(SEQID_LENGTH, sizeof(char));
  char seq[1024];
  char qual[1024];
  int seq_num = 0;

  clock_t start_clock, end_clock;
  if (is_verbose) { start_clock = clock(); }

  strcpy(seqid, "");
  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse id */
    if (sscanf(line, "@%s", seqid) == 1) {

      /* parse the sequence line */
      if (fgetline(stdin, line, sizeof(line)) <= 0) {
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }
      sscanf(line, "%s", seq);

      /* parse the 2nd sequence ID line */
      if (fgetline(stdin, line, sizeof(line)) <= 0 ||
	  sscanf(line, "+%s", seqid2) != 1 ||
	  strcmp(seqid2, seqid) != 0)
	{
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }

      /* parse the quality line */
      if (fgetline(stdin, line, sizeof(line)) <= 0) {
	fprintf(stderr, "bad fastq format\n");
	return -1;
      }
      sscanf(line, "%s", qual);

      /* initialize the sequence length */
      if (seq_length < 0) {
	if (max_seq_length < 0) {
	  seq_length = strlen(seq);
	} else {
	  seq_length = max_seq_length;
	}

	/* create the sequence dataset and write buffer */
        did_seq = _create_sequence_dataset(gid_seqs, seq_length);
	seq_buf = (char*)calloc(seq_buf_size * seq_length, sizeof(char));

	/* create the quality dataset and write buffer */
        did_qual = _create_quality_dataset(gid_seqs, seq_length);
	qual_buf = (char*)calloc(seq_buf_size * seq_length, sizeof(char));
      }

      /* Verify seq and qual length match, else fail */
      if (strlen(seq) != strlen(qual)) {
	fprintf(stderr, "Error: read sequence and quality strings have mismatched length:\n@%s\n%s\n+%s\n%s\n\n", seqid, seq, seqid, qual);
	return -1;
      }

      /* append the seqid to dataset, via a buffer */
      int seqid_buf_loc = seq_num % seqid_buf_size;
      memcpy(seqid_buf + seqid_buf_loc * SEQID_LENGTH, seqid, SEQID_LENGTH);
      if (seqid_buf_loc == seqid_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_seqid, seq_num + 1, seqid_buf, seqid_buf_size, SEQID_LENGTH);
      }
      
      /* append the sequence to dataset, via a buffer */
      int buf_loc = seq_num % seq_buf_size;
      memcpy(seq_buf + buf_loc * seq_length, seq, seq_length);
      if (buf_loc == seq_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_seq, seq_num + 1, seq_buf, seq_buf_size, seq_length);
      }
      
      /* append the quality to dataset, via a buffer - synced with seq */
      memcpy(qual_buf + buf_loc * seq_length, qual, seq_length);
      if (buf_loc == seq_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_qual, seq_num + 1, qual_buf, seq_buf_size, seq_length);
      }
      
      ++seq_num;
      
      if (is_verbose) {
	if (seq_num % 1000000 == 0) {
	  end_clock = clock();
	  double elapsed = ((double)(end_clock - start_clock)) / CLOCKS_PER_SEC;
	  start_clock = end_clock;
	  fprintf(stderr, "Processed %d sequences in %g seconds\n", seq_num, elapsed);
	}
      }
    } 
    else {
      fprintf(stderr, "bad fastq format\n");
      return -1;
    }
  }

  /* flush the seqid buffer to dataset */
  int seqid_buf_loc = seq_num % seqid_buf_size;
  H5UTLwrite_2D_char_buf(did_seqid, seq_num, seqid_buf, seqid_buf_loc, SEQID_LENGTH);
      
  /* flush the seq and qual buffer to dataset */
  int buf_loc = seq_num % seq_buf_size;
  H5UTLwrite_2D_char_buf(did_seq, seq_num, seq_buf, buf_loc, seq_length);
  H5UTLwrite_2D_char_buf(did_qual, seq_num, qual_buf, buf_loc, seq_length);

  if (did_seq) { H5Dclose(did_seq); }
  if (did_qual) { H5Dclose(did_qual); }

  if (is_verbose) { fprintf(stderr, "Processed %d sequences\n", seq_num); }

  return 0;
}


/* initialize the buffer used for writing alignment hits */
void H5BIOinit_hit_buf(hit_buf_t *hit_buf, hid_t did_aln, hid_t tid_hit) {

  hit_buf->did_aln = did_aln;
  hit_buf->hit_num = 0;
  hit_buf->buf_size = HDF_RW_BUF_SIZE;
  /* note no matching free for this calloc; used in short-lived processes */
  hit_buf->buf = (hit_t *)calloc(hit_buf->buf_size, sizeof(hit_t));
  hit_buf->tid_hit = tid_hit;
}


/* write an alignment hit via the buffer */
void H5BIOwrite_hit(hit_buf_t *hit_buf, hit_t *hit) {

  int buf_loc = hit_buf->hit_num % hit_buf->buf_size;
  memcpy(hit_buf->buf + buf_loc, hit, sizeof(hit_t));
  ++hit_buf->hit_num;
  if (buf_loc == hit_buf->buf_size - 1) {
    _write_hit_buf(hit_buf, hit_buf->buf_size);
  }
}


/* flush the bit buffer used during writing */
void H5BIOflush_hit_buf(hit_buf_t *hit_buf) {

  int buf_loc = hit_buf->hit_num % hit_buf->buf_size;
  if (buf_loc > 0) {
    _write_hit_buf(hit_buf, buf_loc);
  }
}


/* write alignment hit buffer into the dataset */
void _write_hit_buf(hit_buf_t *hit_buf, int hit_buf_num) {
  write_hit_buf(hit_buf->did_aln, hit_buf->hit_num, hit_buf->buf, hit_buf_num, hit_buf->tid_hit);
}

/* RFTR - make private / merge into _write_hit_buf above once h5_filter_alignment.c refactored */
void write_hit_buf(hid_t did_aln, int hit_num, const hit_t* hit_buf, int hit_buf_num, hid_t tid_hit) {

  hsize_t dims[1] = { hit_num };
  H5Dextend(did_aln, dims);
  hid_t sid_aln = H5Dget_space(did_aln);
  
  hsize_t slab_start[1] = { hit_num - hit_buf_num };
  hsize_t slab_count[1] = { hit_buf_num };
  H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dwrite(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
  H5Sclose(mid);
}


int compare_refmap(const void *a, const void *b) {
    refmap_t *rma = (refmap_t *)a;
    refmap_t *rmb = (refmap_t *)b;
    int cmp_ref_id;
    int cmp_position;
    int cmp_length;
    
    /* Elements are sorted by reference ID ("gene") first */
    cmp_ref_id = (int)(rma->ref_id - rmb->ref_id);
    if (cmp_ref_id != 0) {
        return cmp_ref_id;
    }
    
    /* If they refer to the same reference ID, then they are sorted
       by the start index (position) */
    cmp_position = (int)(rma->position - rmb->position);
    if (cmp_position != 0) {
        return cmp_position;
    }
    
    /* If they have the same start position then they are sorted by
       length in descending order.  The NCList algorithm requires this
       so don't change it.
      
       The a/b order is reversed since we want to sort by descending
       length. */
    cmp_length = (int)(rmb->length - rma->length);
    return cmp_length;
}


/* create the refmap memory datatype */
hid_t H5BIOhit_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(hit_t));

  H5Tinsert(tid, "read_id",  HOFFSET(hit_t, read_id),  H5T_NATIVE_INT);
  H5Tinsert(tid, "ref_id",   HOFFSET(hit_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "position", HOFFSET(hit_t, position), H5T_NATIVE_INT);
  H5Tinsert(tid, "length",   HOFFSET(hit_t, length),   H5T_NATIVE_INT);
  H5Tinsert(tid, "strand",   HOFFSET(hit_t, strand),   H5T_NATIVE_INT);
  H5Tinsert(tid, "mismatch", HOFFSET(hit_t, mismatch), H5T_NATIVE_INT);
  H5Tinsert(tid, "quality",  HOFFSET(hit_t, quality),  H5T_NATIVE_INT);

  return tid;
}


/* create the refmap memory datatype */
hid_t H5BIOrefmap_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(refmap_t));

  H5Tinsert(tid, "read_id",  HOFFSET(refmap_t, read_id),  H5T_NATIVE_INT);
  H5Tinsert(tid, "ref_id",   HOFFSET(refmap_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "position", HOFFSET(refmap_t, position), H5T_NATIVE_INT);
  H5Tinsert(tid, "length",   HOFFSET(refmap_t, length),   H5T_NATIVE_INT);
  H5Tinsert(tid, "strand",   HOFFSET(refmap_t, strand),   H5T_NATIVE_INT);
  H5Tinsert(tid, "mismatch", HOFFSET(refmap_t, mismatch), H5T_NATIVE_INT);
  H5Tinsert(tid, "quality",  HOFFSET(refmap_t, quality),  H5T_NATIVE_INT);
  H5Tinsert(tid, "hit_id",   HOFFSET(refmap_t, hit_id),   H5T_NATIVE_INT);

  return tid;
}


/* create the NCList level datatype */
hid_t H5BIOnclist_level_type() {

    hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(nclist_level_t));

    H5Tinsert(tid, "start_idx",  HOFFSET(nclist_level_t, start_idx),  H5T_NATIVE_HSIZE);
    H5Tinsert(tid, "end_idx",    HOFFSET(nclist_level_t, end_idx),    H5T_NATIVE_HSIZE);

    return tid;
}


hid_t H5BIOindex_sort_method_enum_type()
{
    hid_t tid;
    unsigned int value;

    tid = H5Tenum_create(H5T_NATIVE_UINT);
    value = 0;
    H5Tenum_insert(tid, "unsorted", &value);
    value = 1;
    H5Tenum_insert(tid, "linear", &value);
    value = 2;
    H5Tenum_insert(tid, "nclist", &value);

    return tid;
}


/* create the cluster memory datatype */
hid_t H5BIOcluster_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(cluster_t));

  H5Tinsert(tid, "ref_id",   HOFFSET(cluster_t, ref_id),   H5T_NATIVE_INT);
  H5Tinsert(tid, "beg_pos",  HOFFSET(cluster_t, beg_pos),  H5T_NATIVE_INT);
  H5Tinsert(tid, "end_pos",  HOFFSET(cluster_t, end_pos),  H5T_NATIVE_INT);
  H5Tinsert(tid, "num_read", HOFFSET(cluster_t, num_read), H5T_NATIVE_INT);

  return tid;
}


/* create the reference sequence index memory datatype */
hid_t H5BIOrefidx_type() {

  hid_t tid = H5Tcreate(H5T_COMPOUND, sizeof(refidx_t));

  H5Tinsert(tid, "offset",   HOFFSET(refidx_t, offset),   H5T_NATIVE_LLONG);
  H5Tinsert(tid, "length",   HOFFSET(refidx_t, length),   H5T_NATIVE_INT);

  return tid;
}


int H5BIOreadmask_alignment(hid_t fid, hid_t gid_aln) {

  /* nothing to do if readmask already exists */
  if (H5UTLdataset_exists(gid_aln, READMASK_DATASET)) {
    return 0;
  }

  hid_t did_aln;
  HDFRET(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));

  /* reads must be imported before alignment readmask computed */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before computing alignment readmask\n", SEQUENCES_GROUP);
    return -1;
  }

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFRET(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFRET(sid_seq = H5Dget_space(did_seq));
  int num_seq, seq_len;
  H5UTLget_2D_extent(sid_seq, &num_seq, &seq_len);
  H5Sclose(sid_seq);
  H5Dclose(did_seq);

  /* initialize the readmask dataset */
  char *readmask = (char *)calloc(num_seq, sizeof(char));

  /* iterate over the alignment */
  hid_t sid_aln;
  HDFRET(sid_aln = H5Dget_space(did_aln));
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(buf_size, sizeof(hit_t));

  /* create the hit memory datatype */
  hid_t tid_hit;
  HDFRET(tid_hit = H5BIOhit_type());

  /* iterate over all hits */
  int i;
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);
    }
    item = hit_buf + i - buf_offset;

    readmask[item->read_id] = 1;
  }

  /* store the read mask */
  hid_t did_mask;
  HDFRET(did_mask = H5UTLcreate_1D_dataset_size(gid_aln, READMASK_DATASET, H5T_NATIVE_CHAR, num_seq));
  hsize_t dims[1] = { num_seq };
  H5Dextend(did_mask, dims);
  H5Dwrite(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, readmask);

  return 0;
}


int H5BIOindex_alignment(hid_t fid, hid_t gid_aln) {

  /* Delete existing datasets */
  if (H5UTLdataset_exists(gid_aln, REFMAP_DATASET)) {
    H5Ldelete(gid_aln, REFMAP_DATASET, H5P_DEFAULT);
    if (H5UTLdataset_exists(gid_aln, ALNINDEL_I_DATASET)) {
      H5Ldelete(gid_aln, ALNINDEL_I_DATASET, H5P_DEFAULT);
    }
    if (H5UTLdataset_exists(gid_aln, SEQUENCE_I_DATASET)) {
      H5Ldelete(gid_aln, SEQUENCE_I_DATASET, H5P_DEFAULT);
    }
    if (H5UTLdataset_exists(gid_aln, QUALITY_I_DATASET)) {
      H5Ldelete(gid_aln, QUALITY_I_DATASET, H5P_DEFAULT);
    }
  }

  hid_t tid_hit, tid_refmap, did_aln, did_map, sid_aln;

  /* create the hit memory datatype */
  HDFRET(tid_hit = H5BIOhit_type());

  /* create the refmap memory datatype */
  HDFRET(tid_refmap = H5BIOrefmap_type());

  /* open the "alignment" dataset (list of individual read/ref hits) */
  HDFRET(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));

  /* create the index dataset 
     (list of hits sorted on numeric ref_id then on hit position) */
  HDFRET(did_map = H5UTLcreate_1D_dataset(gid_aln, REFMAP_DATASET, tid_refmap));

  /* iterate through the alignment dataset, creating index to be sorted */
  HDFRET(sid_aln = H5Dget_space(did_aln));
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(buf_size, sizeof(hit_t));

  /* an index buffer to be sorted later */
  refmap_t *refmap_buf = (refmap_t *)calloc(hit_num, sizeof(refmap_t));
  refmap_t *refmap;

  /* iterate over all hits */
  int i;
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0, refmap = refmap_buf; i < hit_num; i++, refmap++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);
    }
    item = hit_buf + i - buf_offset;

    refmap->read_id  = item->read_id;
    refmap->ref_id   = item->ref_id;
    refmap->position = item->position;
    refmap->length   = item->length;
    refmap->strand   = item->strand;
    refmap->mismatch = item->mismatch;
    refmap->quality  = item->quality;
    refmap->hit_id   = i;
  }

  /* sort the refmap buffer, storing to a dataset as an index into hits */
  H5UTLsort_store_1D_dataset(did_map, tid_refmap,
			     refmap_buf, hit_num, sizeof(refmap_t), 
			     compare_refmap);

  /* use sorted hits to identically sort indel dataset, if present */
  if (H5UTLdataset_exists(gid_aln, ALNINDEL_DATASET)) {
    hid_t did_indel, did_indel_i;
    HDFRET(did_indel = H5Dopen(gid_aln, ALNINDEL_DATASET, H5P_DEFAULT));
    char *indel = (char *)calloc(INDEL_MAX_LENGTH * hit_num, sizeof(char));
    H5Dread(did_indel, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, indel);
    H5Dclose(did_indel);
    char *indel_i = (char *)calloc(INDEL_MAX_LENGTH * hit_num, sizeof(char));
    for (i = 0; i < hit_num; i++) {
      memcpy(indel_i + INDEL_MAX_LENGTH * i, 
	     indel + INDEL_MAX_LENGTH * (refmap_buf + i)->hit_id, 
	     INDEL_MAX_LENGTH);
    }
    free(indel);
    HDFRET(did_indel_i = H5UTLcreate_2D_dataset_size(gid_aln, ALNINDEL_I_DATASET, H5T_NATIVE_CHAR, INDEL_MAX_LENGTH, hit_num));
    H5Dwrite(did_indel_i, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, indel_i);
    H5Dclose(did_indel_i);
    free(indel_i);

    /* index sequences and qualities for SAM export */
    /* open the "sequences" group */
    hid_t gid_seq, did_seq, sid_seq, did_seq_i;
    HDFRET(gid_seq = H5Gopen(fid, SEQUENCES_GROUP, H5P_DEFAULT));
    
    /* use sorted hits to identically sort sequence dataset */
    HDFRET(did_seq = H5Dopen(gid_seq, SEQUENCE_DATASET, H5P_DEFAULT));
    HDFRET(sid_seq = H5Dget_space(did_seq));
    int seq_num, seq_len;
    H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);
    H5Sclose(sid_seq);
    char *seq = (char *)calloc(seq_num * seq_len, sizeof(char));
    H5Dread(did_seq, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, seq);
    H5Dclose(did_seq);
    char *seq_i = (char *)calloc(seq_len * hit_num, sizeof(char));
    for (i = 0; i < hit_num; i++) {
      memcpy(seq_i + seq_len * i, 
	     seq + seq_len * (refmap_buf + i)->read_id, 
	     seq_len);
    }
    free(seq);
    HDFRET(did_seq_i = H5UTLcreate_2D_dataset_size(gid_aln, SEQUENCE_I_DATASET, H5T_NATIVE_CHAR, seq_len, hit_num));
    H5Dwrite(did_seq_i, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, seq_i);
    H5Dclose(did_seq_i);
    free(seq_i);
    
    /* use sorted hits to identically sort quality dataset */
    hid_t did_qual, sid_qual, did_qual_i;
    HDFRET(did_qual = H5Dopen(gid_seq, QUALITY_DATASET, H5P_DEFAULT));
    HDFRET(sid_qual = H5Dget_space(did_qual));
    int qual_num, qual_len;
    H5UTLget_2D_extent(sid_qual, &qual_num, &qual_len);
    H5Sclose(sid_qual);
    char *qual = (char *)calloc(qual_num * qual_len, sizeof(char));
    H5Dread(did_qual, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, qual);
    H5Dclose(did_qual);
    char *qual_i = (char *)calloc(qual_len * hit_num, sizeof(char));
    for (i = 0; i < hit_num; i++) {
      memcpy(qual_i + qual_len * i, 
	     qual + qual_len * (refmap_buf + i)->read_id, 
	     qual_len);
    }
    free(qual);
    HDFRET(did_qual_i = H5UTLcreate_2D_dataset_size(gid_aln, QUALITY_I_DATASET, H5T_NATIVE_CHAR, qual_len, hit_num));
    H5Dwrite(did_qual_i, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, qual_i);
    H5Dclose(did_qual_i);
    free(qual_i);
  }

  return 0;
}


void refmap_item(refmap_t *item, hid_t did, hid_t sid, hid_t tid, hid_t mid, int pp) {

  hsize_t slab_start[1] = { pp };
  hsize_t slab_count[1] = { 1 };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  H5Dread(did, tid, mid, sid, H5P_DEFAULT, item);
}


int refmap_is_match(refmap_t *item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
            
    /* If the hit_id is negative then there is no item available.
     * Fake elements used at the bounds of arrays have a hit_id of -1.
     */
    if (item->hit_id < 0) { 
        return 0; 
    }

    return (
        /* The gene/reference IDs have to be identical */
        item->ref_id == ref_id 
        /* And the end position must be inside the gene bounds */
        && item->position <= end_pos
        /* And the start position must be inside the gene bounds */
        && item->position + item->length - 1 >= start_pos);
}


int refmap_is_first(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
            
    /* Returns true if the current item is a match but the previous one is
     * not.
     * NOTE - only works because reads are identical length. 
     */
    return (
        !refmap_is_match(prior_item, ref_id, start_pos, end_pos, seq_len)
        &&
        refmap_is_match(item, ref_id, start_pos, end_pos, seq_len)
    );
}


int refmap_is_prior(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
    
    /* Returns true if one of the two following critera are met. */
    return (
        /* 1. The current item's gene ID is less than given reference
           ID (reference IDs are sorted in increasing order). */
        item->ref_id < ref_id
        ||
        /* 2. The current item's gene ID is the same as the given reference
           and its end position is BEFORE the given start position. */
        (
            item->ref_id == ref_id 
            && 
            item->position + item->length - 1 < start_pos)
        );
}


int refmap_is_after(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len) {
            
    /* Returns true if one of the following three conditions are true */
    return (
        /* 1. The current item's gene ID is greater than given reference
           ID (reference IDs are sorted in increasing order). */
        item->ref_id > ref_id
        ||
        /* 2. The current item's gene ID is the same as the given reference
           and its start position is AFTER the given end position. */
        (
            item->ref_id == ref_id 
            && 
            item->position + 1 > end_pos
        )
        ||
        /* 3. Both the item and the prior item "match" to the reference.
           If the prior element and this one match, then clearly the selected
           element is not the first one in the region of interest. */
        (
            refmap_is_match(item, ref_id, start_pos, end_pos, seq_len)
            &&
            refmap_is_match(prior_item, ref_id, start_pos, end_pos, seq_len)
        )
    );
}


/* binary search to find matching range, then linear pass over range */
void index_binary_search(hid_t gid_aln, const char *aln_name, 
			 int strand, int ref_id, int start_pos, int end_pos,
			 int seq_len, 
			 void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
			 void *data) {
  hsize_t slab_count[1] = { 1 };
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  hid_t tid = H5BIOrefmap_type();
  hid_t did = H5Dopen(gid_aln, REFMAP_DATASET, H5P_DEFAULT);
  hid_t sid = H5Dget_space(did);

  refmap_t *prior_item = (refmap_t *)malloc(sizeof(refmap_t));
  refmap_t *item = (refmap_t *)malloc(sizeof(refmap_t));

  int ll = -1;
  int rr = H5UTLget_1D_extent(sid);
  int pp;

  int found = 0;
  while ((pp = (rr - ll) / 2) > 0) {
    pp = ll + pp;
    if (pp > 0) {
      refmap_item(prior_item, did, sid, tid, mid, pp - 1);
    } else {
      prior_item->hit_id = -1;
    }
    refmap_item(item, did, sid, tid, mid, pp);
    if (refmap_is_first(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      found = 1;
      break;
    } else if (refmap_is_prior(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      ll = pp;
    } else if (refmap_is_after(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
      rr = pp;
    } else {
      fprintf(stderr, "inconsistent binary search\n");
      exit(1);
    }
  }

  /* tidy up */
  H5Tclose(tid);
  H5Sclose(mid);
  H5Sclose(sid);
  H5Sclose(did);

  free(prior_item);
  free(item);

  if (found) {
    /* linear pass, printing matching range */
    index_linear_search(gid_aln, aln_name, 
			strand, ref_id, start_pos, end_pos, pp, seq_len,
			index_action, data);
  }
}


void index_linear_search(hid_t gid_aln, const char *aln_name, 
			 int strand, int ref_id, int start_pos, int end_pos,
			 int start_id, int seq_len, 
			 void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
			 void *data) {

  hid_t did_refmap = H5Dopen(gid_aln, REFMAP_DATASET, H5P_DEFAULT);
  hid_t sid_refmap = H5Dget_space(did_refmap);
  hid_t tid_refmap = H5BIOrefmap_type();

  hid_t did_indel_i = H5Dopen(gid_aln, ALNINDEL_I_DATASET, H5P_DEFAULT);
  hid_t did_seq_i = H5Dopen(gid_aln, SEQUENCE_I_DATASET, H5P_DEFAULT);
  hid_t did_qual_i = H5Dopen(gid_aln, QUALITY_I_DATASET, H5P_DEFAULT);

  if (did_indel_i < 0 ||
      did_seq_i < 0 ||
      did_indel_i < 0) {
    fprintf(stderr, "indexed read sequences and qualities not available, please index with option --sam\n");
    return;
  }

  hid_t sid_indel_i = H5Dget_space(did_indel_i);
  hid_t sid_seq_i = H5Dget_space(did_seq_i);
  hid_t sid_qual_i = H5Dget_space(did_qual_i);

  int hit_num = H5UTLget_1D_extent(sid_refmap);

  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *refmap_buf  = (refmap_t *)calloc(buf_size, sizeof(refmap_t));
  char     *indel_i_buf = (char     *)calloc(buf_size * INDEL_MAX_LENGTH, sizeof(char));
  char     *seq_i_buf   = (char     *)calloc(buf_size * seq_len, sizeof(char));
  char     *qual_i_buf  = (char     *)calloc(buf_size * seq_len, sizeof(char));

  /* iterate over all */
  int i;
  refmap_t *refmap_item;
  char *indel_item;
  char *seq_item = (char *)calloc(buf_size * seq_len + 1, sizeof(char));
  char *qual_item = (char *)calloc(buf_size * seq_len + 1, sizeof(char));
  int found = 0;
  int buf_offset = start_id - buf_size;
  for (i = start_id; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }

      hsize_t slab_refmap_start[1] = { buf_offset };
      hsize_t slab_refmap_count[1] = { buf_fill };
      hid_t mid_refmap = H5Screate_simple(1, slab_refmap_count, slab_refmap_count); 
      H5Sselect_hyperslab(sid_refmap, H5S_SELECT_SET, slab_refmap_start, NULL, slab_refmap_count, NULL);
      H5Dread(did_refmap, tid_refmap, mid_refmap, sid_refmap, H5P_DEFAULT, refmap_buf);
      H5Sclose(mid_refmap);

      hsize_t slab_indel_i_start[2] = { buf_offset, 0 };
      hsize_t slab_indel_i_count[2] = { buf_fill, INDEL_MAX_LENGTH };
      hid_t mid_indel_i = H5Screate_simple(2, slab_indel_i_count, slab_indel_i_count); 
      H5Sselect_hyperslab(sid_indel_i, H5S_SELECT_SET, slab_indel_i_start, NULL, slab_indel_i_count, NULL);
      H5Dread(did_indel_i, H5T_NATIVE_CHAR, mid_indel_i, sid_indel_i, H5P_DEFAULT, indel_i_buf);
      H5Sclose(mid_indel_i);

      hsize_t slab_seq_i_start[2] = { buf_offset, 0 };
      hsize_t slab_seq_i_count[2] = { buf_fill, seq_len };
      hid_t mid_seq_i = H5Screate_simple(2, slab_seq_i_count, slab_seq_i_count); 
      H5Sselect_hyperslab(sid_seq_i, H5S_SELECT_SET, slab_seq_i_start, NULL, slab_seq_i_count, NULL);
      H5Dread(did_seq_i, H5T_NATIVE_CHAR, mid_seq_i, sid_seq_i, H5P_DEFAULT, seq_i_buf);
      H5Sclose(mid_seq_i);

      hsize_t slab_qual_i_start[2] = { buf_offset, 0 };
      hsize_t slab_qual_i_count[2] = { buf_fill, seq_len };
      hid_t mid_qual_i = H5Screate_simple(2, slab_qual_i_count, slab_qual_i_count); 
      H5Sselect_hyperslab(sid_qual_i, H5S_SELECT_SET, slab_qual_i_start, NULL, slab_qual_i_count, NULL);
      H5Dread(did_qual_i, H5T_NATIVE_CHAR, mid_qual_i, sid_qual_i, H5P_DEFAULT, qual_i_buf);
      H5Sclose(mid_qual_i);
    }

    refmap_item = refmap_buf  + i - buf_offset;
    if (refmap_is_match(refmap_item, ref_id, start_pos, end_pos, seq_len)) {
      found = 1;
      if (strand == 0 ||
	  strand == refmap_item->strand) {

	indel_item  = indel_i_buf + (i - buf_offset) * INDEL_MAX_LENGTH;
	memcpy(seq_item, seq_i_buf + (i - buf_offset) * seq_len, seq_len);
	*(seq_item + seq_len) = 0;
	memcpy(qual_item, qual_i_buf + (i - buf_offset) * seq_len, seq_len);
	*(qual_item + seq_len) = 0;

	(*index_action)(aln_name, refmap_item, indel_item, seq_item, qual_item, start_pos, end_pos, data);
      }
    } else if (found == 1 && refmap_item->position > end_pos) {
      break;
    }
  }

  /* TODO - close the HDF5 objects, free the heap */
}


/* dump of all hits in an alignment set */
#if 0
void index_dump(hid_t gid_aln, const char *aln_name, int strand, 
		void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
		void *data) {
  
  hid_t sid = H5Dget_space(did);
  int hit_num = H5UTLget_1D_extent(sid);

  hid_t tid = H5BIOrefmap_type();

  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *map_buf = (refmap_t *)malloc(sizeof(refmap_t) * buf_size);

  /* iterate over all */
  int i;
  refmap_t *item;
  int buf_offset = -buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did, tid, mid, sid, H5P_DEFAULT, map_buf);
      H5Sclose(mid);
    }
    item = map_buf + i - buf_offset;
    if (strand == 0 ||
	strand == item->strand) {
      (*refmap_action)(aln_name, item, 0, 0, data);
    }
  }

  H5Tclose(tid);
  H5Sclose(sid);
}
#endif


void index_to_sam(const char *aln_name, 
		  const refmap_t *refmap_item, const char *indel_item, 
		  const char *seq_item, const char *qual_item, 
		  int start_pos, int end_pos, void *data) {
  int flag = 0;
  if (refmap_item->strand == -1) flag |= 0x10;
  int mapq = (refmap_item->quality & 0xff);
  char* mrnm = "*"; /* LATER - paired end */
  int mpos = 0; /* LATER - paired end */
  int isize = 0; /* LATER - paired end */
  char opt[10];
  sprintf(opt, "NM:i:%d", refmap_item->mismatch);
  printf("%d\t%d\t%d\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s\t%s\n",
	 refmap_item->read_id,
	 flag,
	 refmap_item->ref_id,
	 refmap_item->position + 1, /* to 1-based */
	 mapq,
	 indel_item,
	 mrnm,
	 mpos,
	 isize, 
	 seq_item,
	 qual_item,
	 opt);
}




/* TODO - search using the bounding read length (seq_len) but filter during
   the linear pass, checking for match on the actual read length */

/* binary search to find matching range, then linear pass over range */
void refmap_binary_search(hid_t did, const char *ref_name, 
			  int strand, int ref_id, int start_pos, int end_pos,
			  int seq_len,  hsize_t refmap_start_idx, hsize_t refmap_end_idx,
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data) {
  
  hsize_t   slab_count[1] = { 1 };
  hid_t     mid = H5Screate_simple(1, slab_count, slab_count); 
  hid_t     tid = H5BIOrefmap_type();
  hid_t     sid = H5Dget_space(did);

  refmap_t *prior_item  = (refmap_t *)calloc(1, sizeof(refmap_t));
  refmap_t *item        = (refmap_t *)calloc(1, sizeof(refmap_t));

  int ll;     /* left search bound */
  int rr;     /* right search bound */
  int pp;     /* probe (element tested) */

  int is_finished = 0;
  int is_first = 0;
  
  
  /***********************************************************************/
  
  
  /* Set up search boundaries */
  ll = refmap_start_idx;
  rr = refmap_end_idx;
  
  /* Binary search to find the first element for a linear scan, if one exists */
  while(!is_finished) {
  
    /* Set the location of the probe */
    pp = ll + ((rr - ll) / 2);
    
    /* Are we done yet? */
    if (rr == ll) {
      is_finished = 1;
    }
  
    /* Copy the probe's data from the dataset to memory */
    refmap_item(item, did, sid, tid, mid, pp);
  
    /* Copy the (probe - 1) data from the dataset to memory */
    /* NOTE: At the 1st index we set the prior ID to -1 to indicate it's not valid. */
    if (pp == refmap_start_idx) {
      prior_item->hit_id = -1;
    } 
    else {
      refmap_item(prior_item, did, sid, tid, mid, pp - 1);
    }
  
    /* TESTS */
    is_first = refmap_is_first(item, prior_item, ref_id, start_pos, end_pos, seq_len);
    
    if(is_first) 
    {
      break;
    }
    else {
    
      /* If that wasn't the first element for the linear scan, we set up for the next round */
    
      if(ll + 1 == rr) {
        /* In the two element case, everything points to the same element (the other one)
         * ensuring algorithm completion.
         */
        if (ll == pp) {
          ll = rr;
        }
        else {
          rr = ll;
        }
        pp = ll;
      }
      else if (refmap_is_prior(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
        /* If the probe is to the left of the target, then search to the right */
        ll = pp;
      }
      else if (refmap_is_after(item, prior_item, ref_id, start_pos, end_pos, seq_len)) {
        /* If the probe is to the right of the target, then search to the left */
        rr = pp;
      }
      else {
        /* Arriving here is an error */
        fprintf(stderr, "inconsistent binary search\n");
        exit(1);
      }
    }
    
  } /* end while */
  
  if (is_first) {
    /* linear pass, doing things with matching range */
    //printf("start pos %d end pos %d  pp %d end idx %d\n", start_pos, end_pos, pp, refmap_end_idx);
    refmap_linear_search(did, ref_name, 
			 strand, ref_id, start_pos, end_pos, pp, refmap_end_idx, seq_len,
			 refmap_action, data);
  }
  
  /* tidy up */
  H5Tclose(tid);
  H5Sclose(mid);
  H5Sclose(sid);

  free(prior_item);
  free(item); 
}


void refmap_linear_search(hid_t did, const char *ref_name, 
			  int strand, int ref_id, int start_pos, int end_pos,
			  int start_id, int end_id, int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data) {


  hid_t sid = H5Dget_space(did);
  int hit_num = hit_num = end_id + 1;

  hid_t tid = H5BIOrefmap_type();

  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *map_buf = (refmap_t *)malloc(sizeof(refmap_t) * buf_size);

  /* iterate over all */
  int i;
  refmap_t *item;
  int found = 0;
  int buf_offset = start_id - buf_size;
  for (i = start_id; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
        buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did, tid, mid, sid, H5P_DEFAULT, map_buf);
      H5Sclose(mid);
    }
    item = map_buf + i - buf_offset;
    if (refmap_is_match(item, ref_id, start_pos, end_pos, seq_len)) {
      found = 1;
      if (strand == 0 || strand == item->strand) {
	      (*refmap_action)(ref_name, item, start_pos, end_pos, data);
      }
    } else if (found == 1 && item->position > end_pos) {
      break;
    }
  }

  H5Tclose(tid);
  H5Sclose(sid);
}


/* dump of all hits in an alignment set */
void refmap_dump(hid_t did, const char *ref_name, int strand, 
		 void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
		 void *data) {

  hid_t sid = H5Dget_space(did);
  int hit_num = H5UTLget_1D_extent(sid);

  hid_t tid = H5BIOrefmap_type();

  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *map_buf = (refmap_t *)malloc(sizeof(refmap_t) * buf_size);

  /* iterate over all */
  int i;
  refmap_t *item;
  int buf_offset = -buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did, tid, mid, sid, H5P_DEFAULT, map_buf);
      H5Sclose(mid);
    }
    item = map_buf + i - buf_offset;
    if (strand == 0 ||
	strand == item->strand) {
      (*refmap_action)(ref_name, item, 0, 0, data);
    }
  }

  H5Tclose(tid);
  H5Sclose(sid);
}


/* TODO - gff 'source' should be alignment algorithm from HDF attribute */
void refmap_to_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  printf("%d\tbiohdf\tread\t%d\t%d\t.\t%s\t.\tID \"%d\"; Name \"%d\"; mm \"%d\"\n", 
	 item->ref_id, 
	 item->position + 1, 
	 item->position + item->length, 
	 item->strand > 0 ? "+" : "-", 
	 item->hit_id, 
	 item->read_id, 
	 item->mismatch);
}


void refmap_to_annot_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {

  hid_t did_annot = *((hid_t *)data);
  char buf[ANNOTATION_LENGTH];
  char score[80];
  H5UTLread_2D_char_line(did_annot, item->hit_id, buf, ANNOTATION_LENGTH);

  /* assumes exact format 'feature "<the_feature>"; <attr>' */
  char *psplit = strchr(buf, ';');
  *(psplit - 1) = '\0';
  char *feature = buf + 9;
  char *attr = psplit + 2;
  /* test for and extract optional score 'score "<the_score>"; <attr>' */
  strcpy(score, ".");
  if (strncmp(attr, "score", 5) == 0) {
    char *psplit2 = strchr(attr, ';');
    *(psplit2 - 1) = '\0';
    strcpy(score, attr + 7);
    attr = psplit2 + 2;
  }
  printf("%d\tbiohdf\t%s\t%d\t%d\t%s\t%s\t.\t%s\n", 
	 item->ref_id, feature, item->position + 1, item->position + item->length, score, item->strand > 0 ? "+" : "-", attr);
}


void refmap_to_bowtie(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  printf("%d%c:<%d,%d,%d>\n", 
	 item->read_id, 
	 item->strand > 0 ? '+' : '-', 
	 item->ref_id, 
	 item->position, 
	 item->mismatch);
}


void refmap_to_bowtie_extended(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  printf("%d%c:<%d,%d,%d,%d>\n", 
	 item->read_id, 
	 item->strand > 0 ? '+' : '-', 
	 item->ref_id, 
	 item->position, 
	 item->mismatch,
	 item->length);
}


void refmap_to_depth(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  int *depth = (int *)data;
  int width = end_pos - start_pos + 1;
  int i;
  for (i = item->position - start_pos; i < item->position - start_pos + item->length; i++) {
    if (i >= 0 && i < width) {
      ++depth[i];
    }
  }
}

void refmap_to_density(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  int *density = (int *)data;
  int i;
  for (i = item->position; i < item->position + item->length; i++) {
    if (i < 0 || i >= 100000) {
      fprintf(stderr, "problem\n");
      exit(1);
    }
    ++density[i];
  }
}

void refmap_to_readmask(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data) {
  char *readmask = (char *)data;
  readmask[item->read_id] = 0;
}


/* write refidx entry into the dataset */
void H5BIOwrite_refidx(hid_t did, int num, const refidx_t* buf, int buf_num, hid_t tid) {

  hsize_t dims[1] = { num };
  H5Dextend(did, dims);
  hid_t sid = H5Dget_space(did);
  
  hsize_t slab_start[1] = { num - buf_num };
  hsize_t slab_count[1] = { buf_num };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dwrite(did, tid, mid, sid, H5P_DEFAULT, buf);
  H5Sclose(mid);
}


int H5BIOcompare_cluster(const void *a, const void *b) {
  cluster_t *cla = (cluster_t *)a;
  cluster_t *clb = (cluster_t *)b;
  return clb->num_read - cla->num_read;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

#define SEQUENCE_RANK 2
int _create_sequence_dataset(hid_t gid, int seq_length) {
  int seq_num = 0;
  hsize_t dims[SEQUENCE_RANK] = { seq_num, seq_length };
  hsize_t maxdims[SEQUENCE_RANK] = { H5S_UNLIMITED, seq_length };
  hid_t sid_seq = H5Screate_simple(SEQUENCE_RANK, dims, maxdims);
  hid_t pid_seq = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[SEQUENCE_RANK] = { 1024, seq_length };
  H5Pset_chunk(pid_seq, SEQUENCE_RANK, chunk_size);
  H5Pset_deflate(pid_seq, HDF_COMPRESSION_LEVEL);
  hid_t did_seq = H5Dcreate(gid, SEQUENCE_DATASET, H5T_NATIVE_CHAR, 
			    sid_seq, H5P_DEFAULT, pid_seq, H5P_DEFAULT);
  H5Pclose(pid_seq);

  return did_seq;
}


#define QUALITY_RANK 2
int _create_quality_dataset(hid_t gid, int length) {
  int num = 0;
  hsize_t dims[QUALITY_RANK] = { num, length };
  hsize_t maxdims[QUALITY_RANK] = { H5S_UNLIMITED, length };
  hid_t sid = H5Screate_simple(QUALITY_RANK, dims, maxdims);
  hid_t pid = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[QUALITY_RANK] = { 1024, length };
  H5Pset_chunk(pid, QUALITY_RANK, chunk_size);
  H5Pset_deflate(pid, HDF_COMPRESSION_LEVEL);
  hid_t did = H5Dcreate(gid, QUALITY_DATASET, H5T_NATIVE_CHAR, 
			sid, H5P_DEFAULT, pid, H5P_DEFAULT);
  H5Pclose(pid);

  return did;
}


#define SEQID_RANK 2
int _create_seqid_dataset(hid_t gid, int seqid_length) {
  int seqid_num = 0;
  hsize_t dims[SEQID_RANK] = { seqid_num, seqid_length };
  hsize_t maxdims[SEQID_RANK] = { H5S_UNLIMITED, seqid_length };
  hid_t sid_seqid = H5Screate_simple(SEQID_RANK, dims, maxdims);
  hid_t pid_seqid = H5Pcreate(H5P_DATASET_CREATE);
  hsize_t chunk_size[SEQID_RANK] = { 1024, seqid_length };
  H5Pset_chunk(pid_seqid, SEQID_RANK, chunk_size);
  H5Pset_deflate(pid_seqid, HDF_COMPRESSION_LEVEL);
  hid_t did_seqid = H5Dcreate(gid, SEQID_DATASET, H5T_NATIVE_CHAR, 
			      sid_seqid, H5P_DEFAULT, pid_seqid, H5P_DEFAULT);
  H5Pclose(pid_seqid);

  return did_seqid;
}

/***************
 * NCList code *
 ************************************************************************/

/**
 * Determines if a subsequent alignment hit is "contained" in the
 * preceeding hit.  Contained means that both endpoints are equal to or
 * inside that of the preceeding hit.  Hits where both endpoints are the
 * same are not contained since they won't cause errors in the search
 * algorithm.
 *
 * @param first     The preceeding element
 * @param second    The subsequent element
 *
 * @returns         1 if the second element is contained in the first.
 *                  0 otherwise.
 */
int _is_contained(refmap_t first, refmap_t second) {

    int end, end2;
    
    /* Containment only applies to identical reference sequences. */
    if(first.ref_id != second.ref_id) return 0;

    end = first.position + first.length - 1;
    end2 = second.position + second.length - 1;

    /* The array should be sorted by position and length so the second
     * element is "contained" if:
     * 1) It's endpoint is inside (<) the first one.
     * 2) The endpoints are the same but the second is inside (>) the
     *    first.
     * If the elements are of equal length we don't care.
     */
    if(end2 < end || (end2 == end && second.position > first.position)) {
        return 1;
    }
    else {
        return 0;
    }

}


/**
 * Helper function that does the work of recursively creating the NCList
 * index.
 * 
 * @param alignments    The unsorted index data array (input).
 * @param n_alignments  The size of the alignments array.
 * @param index         The NCList sorted data (OUT).
 * @param levels        The NCList levels data.  Each entry gives the start and end
 *                          index of each NCList level (OUT).
 * @param n_levels      The size of the levels array (OUT).
 * @param max_levels    The maximum size of the levels array, which can grow (OUT).
 * @param index_offset  The offset into the index array. Used to set the start index
 *                          for each level.
 */
 static void
_nclist_sort_helper(refmap_t *alignments, hsize_t n_alignments,
                    refmap_t *index /*OUT*/,
                    nclist_level_t **levels /*OUT*/, hsize_t *n_levels /*OUT*/,
                    hsize_t *max_levels /*OUT*/, hsize_t *index_offset /*OUT*/) {

    refmap_t *contained;            /* An array containing alignment hits
                                        which are contained in the previous
                                        hit and thus not written out yet. */

    refmap_t *contained_ptr;        /* A pointer into the contained array */
    refmap_t *ptr;                  /* Array walk pointer */
    refmap_t *ptr2;                 /* Array walk pointer */
    
    nclist_level_t *level_ptr;


    hsize_t i;                      /* Iterator for input array walk */
    hsize_t n_copied;               /* The "size" of the index array.
                                     * Really the # of things copied to the index.
                                     */
    hsize_t contained_size;         /* The size of the contained array */
    hsize_t max_contained_size;     /* Max size of the contained array */

    /**********************************************************************/

    /* Recursion base case */
    if (0 == n_alignments) { return; }
    
    /* Alias this so the code is cleaner.
     */
    level_ptr = *levels + *n_levels;
    

    /* If size is 1 then there's nothing to sort.  Just copy the single
     *  element and make a level entry.
     */
    if (1 == n_alignments) {
        level_ptr->start_idx = *index_offset;
        level_ptr->end_idx   = *index_offset;
        memcpy(index, alignments, sizeof(refmap_t));
        (*n_levels)++;
        return;
    }

    /* The contained array contains elements that are removed because they
     * were contained by previous elements. It's set to a small size and
     * grows as necessary.
     */
    max_contained_size = 4096;
    contained = (refmap_t *)calloc(max_contained_size, sizeof(refmap_t));
    contained_ptr = contained;
    contained_size = 0;

    /* ptr and ptr2 walk the alignments array and look for containment. 
     * We're guaranteed at least two alignments since we test for zero and 1
     * in the above code
     */
    ptr = alignments;
    ptr2 = ++alignments;

    /* Walk the input array, copying uncontained elements to the index array
     * and contained elements to the contained array.
     * We inspect one fewer element than a loop would normally test (< n-1)
     * since the last element will be pointed to by the second walk pointer.
     */
    n_copied = 0;
    for (i = 0; i < n_alignments - 1; i++) {

        if(_is_contained(*ptr, *ptr2)) {

            /* CONTAINED - add the contained element to the contained array.
             * Note that the array can grow here if it's too small for the
             * elements.
             */
            if(contained_size + 1 == max_contained_size) {
                max_contained_size *= 2;
                contained = (refmap_t *)realloc(contained, 
                               max_contained_size * sizeof(refmap_t));
                contained_ptr = contained + contained_size;
            }
            memcpy(contained_ptr, ptr2, sizeof(refmap_t));
            contained_size++;
            contained_ptr++;

            /* On the last round we have to copy the FIRST element to the
             * index as well.
             */
            if (i == n_alignments - 2) {
                memcpy(index, ptr, sizeof(refmap_t));
                n_copied++;
                index++;
            }

            ptr2++;
        }
        else {

            /* NOT CONTAINED - add the base/first element to the index array. */
            memcpy(index, ptr, sizeof(refmap_t));
            n_copied++;
            index++;

            /* On the last round we have to copy the LAST element in as well. */
            if (i == n_alignments - 2) {
                memcpy(index, ptr2, sizeof(refmap_t));
                n_copied++;
                index++;
            }

            ptr = ptr2;
            ptr2++;
        }
    }

    /* Adjust the bounds list. It grows automatically when necessary */
    if(*n_levels + 1 == *max_levels) {
        *max_levels *= 2;
        *levels = (nclist_level_t *)realloc(*levels, 
                                    *max_levels * sizeof(nclist_level_t));
        level_ptr = *levels + *n_levels;
    }
    level_ptr->start_idx    = *index_offset;
    level_ptr->end_idx      = *index_offset + n_copied - 1;
    
    /* Recursively create an NCList out of the elements that contained others. */
    *index_offset = *index_offset + n_copied;
    (*n_levels)++;
    _nclist_sort_helper(contained, contained_size, 
                        index,
                        levels, n_levels, max_levels,
                        index_offset);
    
    free(contained);

    return;
}

/**
 * Sorts an array of refmap_t structs using NCList indexing.  Returns a pointer
 * to the sorted data which you must free() yourself.  The original data will be
 * unchanged.
 * 
 * Called by H5BIOcreate_nclist_index() and not intended for outside use.
 *
 * @param alignments
 * @param n_alignments
 * @param index
 * @param levels
 * @param n_levels
 */
static void
_nclist_sort(refmap_t *alignments, hsize_t n_alignments,
             refmap_t *index /*OUT*/,
             nclist_level_t **levels /*OUT*/, hsize_t *n_levels /*OUT*/) {
    
    hsize_t     max_levels;
    hsize_t     index_offset;   /* Offset into the index array used
                                    in the recursion. */
    
    
    
    /* Create the NCList levels buffer.  It's initially set to a default
       size that should prevent reallocation later.  It will automatically
       grow if needed. */
    max_levels = 1024;
    *n_levels = 0;
    *levels = (nclist_level_t *)calloc(max_levels, sizeof(nclist_level_t));
    
    /* Create the NCList index */
    index_offset = 0;
    _nclist_sort_helper(alignments, n_alignments, index,
                        levels, n_levels, &max_levels, &index_offset);
    
    /* Trim any unused space from the levels array */
    *levels = (nclist_level_t *)realloc(*levels, *n_levels * sizeof(nclist_level_t));
}


void H5BIOcreate_nclist_index(hid_t did_index, hid_t did_nclist_level, 
                              hid_t tid_index, hid_t tid_nclist_level, 
                              refmap_t *alignments, hsize_t n_alignments, 
                              size_t alignment_struct_size, 
                              int(*comparator)(const void *, const void *)) {
    
    refmap_t        *index;         /* The nclist index array */
    nclist_level_t  *levels;        /* The nclist levels array */
    hsize_t         n_levels;       /* The current number of levels stored */ 
    
    hsize_t         dims[1];        /* Used to reset the dataset sizes */
    

    /* Sort the alignments using the stdlib's qsort() and a comparator
     * that is compatible with NCList indexing.
     * FIXME: Assumes that all the alignment hits can fit into memory
     * which will not be true for large collections of data.
     */
    qsort(alignments, n_alignments, alignment_struct_size, comparator);
    
    /* Create the index buffer.  It's the same size as the input alignments
       buffer. */
    index = (refmap_t *)calloc(n_alignments, sizeof(refmap_t));
    
    /* Sort the alignments using an NCList-like algorithm and store
     * them in memory.
     */
    n_levels = 0;
    _nclist_sort(alignments, n_alignments,
                 index /*OUT*/, &levels /*OUT*/, &n_levels /*OUT*/);
            
    /* Fill the NCList index dataset. */
    dims[0] = n_alignments;
    H5Dset_extent(did_index, dims);
    H5Dwrite(did_index, tid_index, H5S_ALL, H5S_ALL, H5P_DEFAULT, index);
    
    /* Fill the NCList levels dataset. */
    dims[0] = n_levels;
    H5Dset_extent(did_nclist_level, dims);
    H5Dwrite(did_nclist_level, tid_nclist_level, H5S_ALL, H5S_ALL, H5P_DEFAULT, levels);
    
    if(index != NULL) free(index);
    if(levels != NULL) free(levels);
}

int H5BIOindex_alignment_nclist(hid_t fid, hid_t gid_aln) {

    hid_t   tid_hit;            /* HDF5 type ID for alignment hits */ 
    hid_t   tid_index;          /* HDF5 type ID for index elements */ 
    hid_t   tid_nclist_level;   /* HDF5 type ID for NCList levels */ 
    
    hid_t   did_hit;            /* HDF5 dataset ID for alignment hits */
    hid_t   did_index;          /* HDF5 dataset ID for index elements */
    hid_t   did_nclist_level;   /* HDF5 dataset ID for NCList levels */
    
    hid_t   sid_hit;            /* HDF5 dataspace ID for alignment hits */
    
    refmap_t *hits;             /* The alignment hits from the BioHDF file */
    hsize_t n_hits;             /* The number of hits in the hits dataset */
    
    /* These are all used in the buffered I/O scheme for reading
     * alignment hits 
     */
    int         i;              /* Iterator */
    hit_t       *hit_buf;       /* The I/O buffer */
    int         hit_buf_size;   /* Size of the I/O buffer */
    hit_t       *buf_ptr;       /* Pointer to the current hit in the buffer */
    refmap_t    *hit_ptr;       /* Pointer to the alignment hit being filled */    
    int buf_offset;
    int buf_fill;
    hsize_t slab_start[1];      /* Hyperslab start and count parameters for */
    hsize_t slab_count[1];      /* filling the hit buffer */
    hid_t mid;                  /* Memory dataspace ID for buffer fill */
    
    /* These variables are used for labeling the index as NCList sorted
     * using an attribute.
     */
    hid_t aid;                  /* Attribute ID for sort type */
    hid_t index_type_tid;       /* Index sort type enum ID */
    unsigned int sort_type;     /* The type of sorting used
                                    (HDF5 enum value)*/
    
    /***********************************************************************/
    
    /* Delete the index dataset if it exists. */
    if (H5UTLdataset_exists(gid_aln, REFMAP_DATASET)) {
        H5Ldelete(gid_aln, REFMAP_DATASET, H5P_DEFAULT);
    }

    /* Delete the nclist level dataset if it exists. */
    if (H5UTLdataset_exists(gid_aln, NCLIST_LEVEL_DATASET)) {
        H5Ldelete(gid_aln, NCLIST_LEVEL_DATASET, H5P_DEFAULT);
    }

    /* Create HDF5 datatypes. */
    HDFRET(tid_hit = H5BIOhit_type());
    HDFRET(tid_index = H5BIOrefmap_type());
    HDFRET(tid_nclist_level = H5BIOnclist_level_type());

    /* Open the alignment dataset and get its size. */
    HDFRET(did_hit = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));
    HDFRET(sid_hit = H5Dget_space(did_hit));
    n_hits = H5UTLget_1D_extent(sid_hit);

    /* Create the index and NCList level datasets. */
    HDFRET(did_index = H5UTLcreate_1D_dataset(gid_aln,
        REFMAP_DATASET, tid_index));
    HDFRET(did_nclist_level = H5UTLcreate_1D_dataset(gid_aln,
        NCLIST_LEVEL_DATASET, tid_nclist_level));
    
    /* Get a chunk of memory for an IO buffer used to read the alignment
     * hits and to store the (currently unsorted) index data.
     */
    hit_buf_size = HDF_RW_BUF_SIZE;
    hit_buf      = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
    hits         = (refmap_t *)calloc(n_hits, sizeof(refmap_t));

    /* Copy the alignment hit data from the alignment hit dataset into the
     * chunk of memory where the index is being stored.  Uses a buffer for
     * more efficient I/O.
     */
    buf_offset = -hit_buf_size;
    for (i = 0, hit_ptr = hits; i < n_hits; i++, hit_ptr++) {
    
        /* If the buffer is empty, refill it */
        if (i >= buf_offset + hit_buf_size) {
            buf_offset += hit_buf_size;
            buf_fill = n_hits - buf_offset;
            if (buf_fill > hit_buf_size) {
                buf_fill = hit_buf_size;
            }
            slab_start[0] = buf_offset;
            slab_count[0] = buf_fill;
            H5Sselect_hyperslab(sid_hit, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
            mid = H5Screate_simple(1, slab_count, slab_count); 
            H5Dread(did_hit, tid_hit, mid, sid_hit, H5P_DEFAULT, hit_buf);
            H5Sclose(mid);
        }
        
        buf_ptr = hit_buf + i - buf_offset;

        hit_ptr->read_id  = buf_ptr->read_id;
        hit_ptr->ref_id   = buf_ptr->ref_id;
        hit_ptr->position = buf_ptr->position;
        hit_ptr->length   = buf_ptr->length;
        hit_ptr->strand   = buf_ptr->strand;
        hit_ptr->mismatch = buf_ptr->mismatch;
        hit_ptr->hit_id   = i;
    }
    free(hit_buf);

    /* Sort and store the index data using NCList indexing. */
    H5BIOcreate_nclist_index(did_index, did_nclist_level,
        tid_index, tid_nclist_level, 
        hits, n_hits, sizeof(refmap_t), compare_refmap);
    free(hits);

    /* Label the index as being NCList sorted using an attribute. */
    index_type_tid = H5BIOindex_sort_method_enum_type();
    aid = H5Acreate2(did_index, "sort_method", index_type_tid,
                    H5Screate(H5S_SCALAR), H5P_DEFAULT, H5P_DEFAULT);
    sort_type = H5Tget_member_index(index_type_tid, "nclist");
    H5Awrite(aid, index_type_tid, &sort_type);
    H5Aclose(aid);
    H5Tclose(index_type_tid);

    /* Clean up HDF5 resources */
    H5Tclose(tid_hit);
    H5Tclose(tid_index);
    H5Tclose(tid_nclist_level);
    H5Sclose(sid_hit);
    H5Dclose(did_hit);
    H5Dclose(did_index);
    H5Dclose(did_nclist_level);
    
    return 0;
}

void H5BIOrefmap_nclist_search(hid_t index_did, hid_t level_did, const char *ref_name, 
            int strand, int ref_id, int start_pos, int end_pos, int seq_len, 
            void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
            void *data) {

    /* Identifiers for the NCList level dataset */
    hid_t   level_sid;              /* NCList level dataspace ID */
    hid_t   level_tid;              /* NCList level type ID */
    hsize_t level_size;             /* The size of the NCList level dataset */

    hsize_t i;                      /* Iterator over the nclist levels */           

    nclist_level_t *level_array;    /* Array for the NCList levels */
    
    /* These are used when there is no NCList level information and we'll
     * be doing a regular binary search. 
     */
    hid_t   index_sid;              /* Index dataspace ID for size info */
    hsize_t no_nclist_start_pos;    /* Used to calculate the start and */
    hsize_t no_nclist_end_pos;      /* end positions */

    /* ------------------------------------------------------------------- */
    
    /* NCList search */
    if(level_did > 0) {

        /* Get NCList level dataspace/set/type IDs and size */
        level_sid   = H5Dget_space(level_did);
        level_size  = H5UTLget_1D_extent(level_sid);
        level_tid   = H5BIOnclist_level_type();

        /* Read in all the NCList level data */
        level_array = (nclist_level_t *)calloc(level_size, sizeof(nclist_level_t));
        H5Dread(level_did, level_tid, H5S_ALL, H5S_ALL, H5P_DEFAULT, level_array);

        /* Iterate over all NCList levels */
        for(i = 0; i < level_size; i++) {

            /* Binary search the data using the NCList level indexes */
            refmap_binary_search(index_did, ref_name, strand, ref_id,
                start_pos, end_pos, seq_len,
                level_array[i].start_idx, level_array[i].end_idx,
                refmap_action, data);
        }
    }
    else {
        index_sid = H5Dget_space(index_did);
        no_nclist_start_pos = 0;
        no_nclist_end_pos = H5UTLget_1D_extent(index_sid);
        
        /* No NCList level dataset so regular binary search */
        refmap_binary_search(index_did, ref_name, strand, ref_id,
            start_pos, end_pos, seq_len,
            no_nclist_start_pos, no_nclist_end_pos,
            refmap_action, data);
    }
}
