/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 *
 * BioHDF library header. 
 *
 *****************************************************************************/

#ifndef BIOHDF_H
#define BIOHDF_H 1

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "hdf5.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* BioHDF command help link */
#define CMD_LINK "For more information about BioHDF visit http://www.biohdf.org/\n"

/* fixed maximum data lengths, with lots of potential for trouble */
#define SEQID_LENGTH  100
#define REFID_LENGTH  1024
#define REFDEF_LENGTH 1024
#define ANNOTATION_LENGTH 1024
#define SEQ_MAX_LENGTH 1024
#define INDEL_MAX_LENGTH 100

#define REFERENCES_GROUP  "references"
#define SEQUENCES_GROUP   "sequences"
#define ALIGNMENTS_GROUP  "alignments"
#define SEQUENCE_DATASET  "sequence"
#define SEQUENCE_I_DATASET  "sequence_i" /* temporary */
#define QUALITY_DATASET   "quality"
#define QUALITY_I_DATASET "quality_i" /* temporary */
#define SEQID_DATASET     "seqid"
#define ALIGNMENT_DATASET "alignment"
#define ALNHEADER_DATASET "header"
#define ALNINDEL_DATASET  "indel"
#define ALNINDEL_I_DATASET  "indel_i" /* temporary */
#define REFMAP_DATASET    "refmap"
#define NCLIST_LEVEL_DATASET    "nclistlevel"
#define READMASK_DATASET  "readmask"
#define CLUSTER_DATASET   "cluster"
#define ANNOTATION_DATASET "annotation"
#define REFID_DATASET     "refid"
#define REFDEF_DATASET    "refdef"
#define REFIDX_DATASET    "refidx"
#define REFSEQ_DATASET    "refseq"


/* basic representation of an alignment hit */
typedef struct _hit_t {
  int read_id;
  int ref_id;
  int position;
  int length;
  int strand;
  int mismatch;
  int quality;
} hit_t;


/* used for buffering alignment hits */
typedef struct _hit_buf_t {
  hid_t did_aln;
  hid_t tid_hit;
  int hit_num;
  int buf_size;
  hit_t *buf;
} hit_buf_t;


/* used for indexing alignment hits, sorted by reference id and position */
typedef struct _refmap_t {
  int read_id;
  int ref_id;
  int position;
  int length;
  int strand;
  int mismatch;
  int quality;
  int hit_id;
} refmap_t;

/* Represents the NCList index levels */
typedef struct _nclist_level_t {
    hsize_t start_idx;
    hsize_t end_idx;
} nclist_level_t;


/* representation of a cluster of aligned reads - lookup into the alignment
   dataset would be required to recover the reads themselves */
typedef struct _cluster_t {
  int ref_id;
  int beg_pos;
  int end_pos;
  int num_read;
} cluster_t;


/* used for indexing into reference sequence 1D-char dataset */
typedef struct _refidx_t {
  long long offset;
  int length;
} refidx_t;


/******************************************************************************
 * Exported function declarations
 *****************************************************************************/

/* parse sequences from fasta format into sequence and seqid datasets */
int H5BIOparse_fastA(hid_t gid_seqs, int max_seq_length, int is_verbose);

/* parse sequences from fastq format into sequence and seqid datasets */
int H5BIOparse_fastQ(hid_t gid_seqs, int max_seq_length, int is_verbose);

/* create the hit memory datatype */
hid_t H5BIOhit_type();

/* create the refmap memory datatype */
hid_t H5BIOrefmap_type();

/* create the nclist level datatype */
hid_t H5BIOnclist_level_type();

/* Create the refmap sort method type */
hid_t H5BIOindex_sort_method_enum_type();

/* create the cluster memory datatype */
hid_t H5BIOcluster_type();

/* create the reference sequence index memory datatype */
hid_t H5BIOrefidx_type();

/* create the readmask dataset for this alignment */
int H5BIOreadmask_alignment(hid_t fid, hid_t gid_aln);

/* create the refmap index dataset for this alignment */
int H5BIOindex_alignment(hid_t fid, hid_t gid_aln);

/* initialize the buffer used for writing alignment hits */
void H5BIOinit_hit_buf(hit_buf_t *hit_buf, hid_t did_aln, hid_t tid_hit);

/* write an alignment hit via the buffer */
void H5BIOwrite_hit(hit_buf_t *hit_buf, hit_t *hit);

/* flush the bit buffer used during writing */
void H5BIOflush_hit_buf(hit_buf_t *hit_buf);


void write_hit_buf(hid_t did_aln, int hit_num, const hit_t* hit_buf, int hit_buf_num, hid_t hit_tid);

int compare_refmap(const void *a, const void *b);

void refmap_item(refmap_t *item, hid_t did, hid_t sid, 
		 hid_t tid, hid_t mid, int pp);

int refmap_is_match(refmap_t *item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_first(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_prior(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

int refmap_is_after(refmap_t *item, refmap_t *prior_item, 
		    int ref_id, int start_pos, int end_pos, int seq_len);

void index_binary_search(hid_t gid_aln, const char *aln_name, 
			 int strand, int ref_id, int start_pos, int end_pos,
			 int seq_len,
			 void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
			 void *data);

void index_linear_search(hid_t gid_aln, const char *aln_name, 
			 int strand, int ref_id, int start_pos, int end_pos,
			 int start_id, int seq_len, 
			 void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
			 void *data);

#if 0
void index_dump(hid_t did, const char *ref_name, int strand, 
		void (*index_action)(const char *, const refmap_t *, const char *, const char *, const char *, int, int, void *),
		void *data);
#endif

void index_to_sam(const char *aln_name, 
		  const refmap_t *refmap_item, const char *indel_item, 
		  const char *seq_item, const char *qual_item, 
		  int start_pos, int end_pos, void *data);

/* binary search into an alignment set */
void refmap_binary_search(hid_t did, const char *ref_name, 
			  int strand, int ref_id, int start_pos, int end_pos,
			  int seq_len, hsize_t refmap_start_idx, hsize_t refmap_end_idx,
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data);

/* linear search over all hits in an alignment set */
void refmap_linear_search(hid_t did, const char *ref_name, 
			  int strand, int ref_id, int start_pos, int end_pos,
			  int start_id,	int end_id, int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data);

/* dump of all hits in an alignment set */
void refmap_dump(hid_t did, const char *ref_name, int strand, 
		 void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
		 void *data);

void refmap_to_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_annot_gff(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_bowtie(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_bowtie_extended(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_depth(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_density(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

void refmap_to_readmask(const char *ref_name, const refmap_t *item, int start_pos, int end_pos, void *data);

/* write refidx entry into the dataset */
void H5BIOwrite_refidx(hid_t did, int num, const refidx_t* buf, int buf_num, hid_t tid);

int H5BIOcompare_cluster(const void *a, const void *b);

/***************
 * NCList code *
 ***************/

/** 
 * Sorts an alignment hit dataset in a BioHDF file using NCList indexing.
 *
 * @param did_index             Dataset ID for the alignment hits.
 * @param did_nclist_level      Dataset ID for the NCList levels.
 * @param tid_index             Type ID for the alignment hits.
 * @param tid_nclist_level      Type ID for the NCList levels.
 * @param alignments            The array of alignments.
 * @param n_alignments          The size of the alignments array.
 * @param alignment_struct_size The size of an alignment struct in bytes.
 * @param comparator            NCList-compatible qsort comparison function.
 */
void H5BIOcreate_nclist_index(hid_t did_index, hid_t did_nclist_level, 
                              hid_t tid_index, hid_t tid_nclist_level, 
                              refmap_t *alignments, hsize_t n_alignments, 
                              size_t alignment_struct_size, 
                              int(*comparator)(const void *, const void *));
                              
/**
 * Creates an index dataset for alignment hit data using NCList-like
 * indexing.
 *
 * @param fid       The HDF5 file id
 * @param gid_aln   The alignment group id which contains the dataset to index
 *
 * @return          An error code.  Failure is indicated by nonzero values.
 */
int H5BIOindex_alignment_nclist(hid_t fid, hid_t gid_aln);

/**
 * Search NCList-sorted alignment hits.  Uses a function pointer to designate
 * the action to take on each found alignment hit (file I/O, statistics, etc.).
 *
 * @param index_did         The dataset ID of the alignment hit NCList
 * @param level_did         The dataset ID of the NCList levels dataset
 * @param ref_name          The 
 * @param strand            Whether we are interested in the + or - strand
 * @param ref_id            The reference ID to match
 * @param start_pos         The start position on the reference ID
 * @param end_pos           The end position on the reference ID
 * @param seq_len           
 * @param refmap_action     The action to take on each found hit (function pointer).
 * @param data              Custom data needed by the refmap_action function
 *
 */
void H5BIOrefmap_nclist_search(hid_t index_did, hid_t level_did, const char *ref_name, 
			  int strand, int ref_id, int start_pos, int end_pos,
			  int seq_len, 
			  void (*refmap_action)(const char *, const refmap_t *, int, int, void *),
			  void *data);

#endif /* BIOHDF_H */
