/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
			  int *overlap, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char aln_name[HDF_OBJNAME_MAX];
  int overlap = 0;
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, aln_name, &overlap, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be imported before alignment annotations added */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* open the alignments group */
  hid_t gid_alns, gid_aln, did_map, sid_map;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));
  HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));
  HDFERR(did_map = H5Dopen(gid_aln, REFMAP_DATASET, H5P_DEFAULT));
  HDFERR(sid_map = H5Dget_space(did_map));
  int refmap_num = H5UTLget_1D_extent(sid_map);

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  hid_t did_seq = -1;
  hid_t sid_seq = -1;
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_len;
  H5UTLget_2D_extent(sid_seq, NULL, &seq_len);
  if (sid_seq >= 0) H5Sclose(sid_seq);
  if (did_seq >= 0) H5Dclose(did_seq);

  /* initialize the cluster dataset - refmap_size is upper limit on clusters */
  cluster_t *cluster = (cluster_t *)calloc(refmap_num, sizeof(cluster_t));
  cluster_t *cl_ptr = cluster;
  int cl_num = 0;

  /* use fixed-size buffer to iterate over the refmap dataset */
  int buf_size = HDF_RW_BUF_SIZE;
  refmap_t *map_buf = (refmap_t *)calloc(buf_size, sizeof(refmap_t));

  /* create the refmap memory datatype */
  hid_t tid_map;
  HDFERR(tid_map = H5BIOrefmap_type());

  /* iterate over all hits */
  int i;
  refmap_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < refmap_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = refmap_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_map, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_map, tid_map, mid, sid_map, H5P_DEFAULT, map_buf);
      H5Sclose(mid);
    }
    item = map_buf + i - buf_offset;

    /* cluster the read alignments */
    if (!(cl_num == 0 && cl_ptr->num_read == 0) &&
	item->ref_id == cl_ptr->ref_id &&
	cl_ptr->end_pos - item->position >= overlap &&
	item->position + item->length - cl_ptr->beg_pos >= overlap) {
      ++cl_ptr->num_read;
      if (cl_ptr->beg_pos > item->position) {
	cl_ptr->beg_pos = item->position;
      }
      if (cl_ptr->end_pos < item->position + item->length) {
	cl_ptr->end_pos = item->position + item->length;
      }
    } else {
      if (cl_num != 0) ++cl_ptr;
      ++cl_num;
      cl_ptr->num_read = 1;
      cl_ptr->ref_id = item->ref_id;
      cl_ptr->beg_pos = item->position;
      cl_ptr->end_pos = item->position + item->length;
    }
  }

  /* sort the clusters by num_read descending */
  qsort(cluster, cl_num, sizeof(cluster_t), H5BIOcompare_cluster);

  /* store the sorted cluster dataset */
  hid_t tid_cl, did_cl;
  HDFERR(tid_cl = H5BIOcluster_type());
  HDFERR(did_cl = H5UTLcreate_1D_dataset_size(gid_aln, CLUSTER_DATASET, tid_cl, cl_num));
  hsize_t dims[1] = { cl_num };
  H5Dextend(did_cl, dims);
  H5Dwrite(did_cl, tid_cl, H5S_ALL, H5S_ALL, H5P_DEFAULT, cluster);


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */
  /* free(hit_buf); */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Form read clusters based on their alignment to a reference.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Read alignment set to be clustered\n"
  "  [--overlap|-v <alignment>]\n"
  "      Overlap distance required to cluster; a negative value implies gaps \n"
  "      between read alignments are permitted without breaking the cluster\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_cluster_alignment -f bvt.h5 -a genome -v -25\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
			  int *overlap, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:v:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"overlap",  required_argument, 0, 'v'},
    {"h5-file",  required_argument, 0, 'f'},
    {"aln",      required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* overlap required for clustering, default 0 */
    case 'v':
      if (optarg == NULL) goto USAGE;
      *overlap = atoi(optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}



