/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *window, char *window_method, int *is_all,
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/


int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char out_fmt[40];
  char aln_name[40];
  int ref_id, ref_beg, ref_end, strand, window;
  char window_method[40];
  int is_all = 0;
  int is_verbose = 0;
  int is_debug   = 0;

  hid_t fid, gid_aln, did;
  hid_t level_did = -1;

  /* Option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");
  strcpy(out_fmt,      "gff");
  ref_id  = 0;
  ref_beg = 1;
  ref_end = 10000;
  strand  = 0;
  window  = 1;
  strcpy(window_method,"maximum");

  /* Parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, out_fmt, aln_name,
			&ref_id, &ref_beg, &ref_end, &strand, 
			&window, window_method, &is_all, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* read from the HDF file */
  HDFERR(fid = H5Fopen(h5_file_name, 
		       H5F_ACC_RDWR, /* RDWR to enable data index creation */
		       H5P_DEFAULT));

  /* open the requested alignment group */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", ALIGNMENTS_GROUP, aln_name);
  HDFERR(gid_aln = H5Gopen(fid, hdf_path, H5P_DEFAULT));

  /* open the refmap dataset */
  HDFERR(did = H5Dopen(gid_aln, REFMAP_DATASET, H5P_DEFAULT));
  
  /* open the NCList level dataset */
  level_did = H5Dopen(gid_aln, NCLIST_LEVEL_DATASET, H5P_DEFAULT);

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  hid_t gid_seq, did_seq, sid_seq;
  HDFERR(gid_seq = H5Gopen(fid, SEQUENCES_GROUP, H5P_DEFAULT));
  HDFERR(did_seq = H5Dopen(gid_seq, SEQUENCE_DATASET, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_len;
  H5UTLget_2D_extent(sid_seq, NULL, &seq_len);
  H5Sclose(sid_seq);

  /* search for reads that align to specified region of a ref_id */
  if (strcmp(out_fmt, "sam") == 0) {
    /* Export SAM header */
    hid_t did_header, sid_header;
    HDFERR(did_header = H5Dopen(gid_aln, ALNHEADER_DATASET, H5P_DEFAULT));
    HDFERR(sid_header = H5Dget_space(did_header));
    int header_size = H5UTLget_1D_extent(sid_header);
    char *header = (char *)calloc(header_size, sizeof(char));
    H5Dread(did_header, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, header);
    HDFERR(H5Dclose(did_header));
    printf("%s", header);
    if (is_all) {
#if 0
      index_dump(gid_aln, aln_name, strand, index_to_sam, NULL);
#endif
    } else {
      index_binary_search(gid_aln, aln_name, 
			  strand, ref_id, ref_beg, ref_end, seq_len,
			  index_to_sam, NULL);
    }
  }
  else if (strcmp(out_fmt, "gff") == 0) {
    printf("track name=hdf_output description=\"HDF Output\"\n");
    if (is_all) {
      refmap_dump(did, aln_name, strand, refmap_to_gff, NULL);
    } else {
      H5BIOrefmap_nclist_search(did, level_did, aln_name, 
			   strand, ref_id, ref_beg, ref_end, seq_len,
			   refmap_to_gff, NULL);
    }
  }
  else if (strcmp(out_fmt, "bowtie") == 0) {
    if (is_all) {
      refmap_dump(did, aln_name, strand, refmap_to_bowtie, NULL);
    } else {
      H5BIOrefmap_nclist_search(did, level_did, aln_name, 
			   strand, ref_id, ref_beg, ref_end, seq_len,
			   refmap_to_bowtie, NULL);
    }
  }
  else if (strcmp(out_fmt, "bowtie_extended") == 0) {
    if (is_all) {
      refmap_dump(did, aln_name, strand, refmap_to_bowtie_extended, NULL);
    } else {
      H5BIOrefmap_nclist_search(did, level_did, aln_name, 
			   strand, ref_id, ref_beg, ref_end, seq_len,
			   refmap_to_bowtie_extended, NULL);
    }
  }
  else if (strcmp(out_fmt, "wig") == 0) {
    int i;
    int width = ref_end - ref_beg + 1;
    int *depth = (int *)calloc(width, sizeof(int));
    int method;
    if (strcmp("max", window_method) == 0 ||
	strcmp("maximum", window_method) == 0) {
      method = 0;
    }
    else if (strcmp("avg", window_method) == 0 ||
	     strcmp("average", window_method) == 0) {
      method = 1;
    }
    else {
      fprintf(stderr, "Unknown window method --window-method %s\n", window_method);
      goto _error;
    }
    printf("track name=hdf_output description=\"HDF Output\"\n");
    H5BIOrefmap_nclist_search(did, level_did, aln_name, 
			 strand, ref_id, ref_beg, ref_end, seq_len,
			 refmap_to_depth, depth);
    if (window > 1) {
      /* output one line per window size */
      /* find aggregate (max or avg) value in window */
      int max_depth = 0;
      int total_depth = 0;
      int window_count = 0;
      printf("fixedStep chrom=%d start=%d step=%d span=%d\n", ref_id, ref_beg, window, window);
      for (i = 0; i < width; i++) {
	if (max_depth < depth[i]) {
	  max_depth = depth[i];
	}
	total_depth += depth[i];
	++window_count;
	if ((i % window) == window - 1) {
	  switch (method) {
	  case 0: printf("%d\n", max_depth); break;
	  case 1: printf("%.2f\n", (float)total_depth / (float)window_count); break;
	  }
	  max_depth = 0;
	  total_depth = 0;
	  window_count = 0;
	}
      }
      if (window_count != 0) {
	switch (method) {
	case 0: printf("%d\n", max_depth); break;
	case 1: printf("%.2f\n", (float)total_depth / (float)window_count); break;
	}
      }
    }
    else {
      /* output one line per base position */
      printf("fixedStep chrom=%d start=%d step=1\n", ref_id, ref_beg);
      for (i = 0; i < width; i++) {
	printf("%d\n", depth[i]);
      }
    }
  }
  else if (strcmp(out_fmt, "read_count") == 0) {
    /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
    /* TODO - library routine anyone? */
    char hdf_path[PATH_MAX];
    sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
    hid_t did_seq, sid_seq;
    HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
    HDFERR(sid_seq = H5Dget_space(did_seq));
    int num_seq, seq_len;
    H5UTLget_2D_extent(sid_seq, &num_seq, &seq_len);
    H5Sclose(sid_seq);
    H5Dclose(did_seq);

    char *readmask = (char *)calloc(num_seq, sizeof(char));
    memset(readmask, 1, num_seq); /* init to 1, set to 0 when found */
    H5BIOrefmap_nclist_search(did, level_did, aln_name, 
			 strand, ref_id, ref_beg, ref_end, seq_len,
			 refmap_to_readmask, readmask);
    /* Compute count of hit reads from readmask */
    int read_count, i;
    for (i = 0, read_count = 0; i < num_seq; i++) {
      if (readmask[i] == 0) { 
	/* TEST printf("%d\n", i); TEST */
	++read_count;
      }
    }
    printf("%d\n", read_count);
  }
  else {
    fprintf(stderr, "Unknown output format --out-fmt %s\n", out_fmt);
    goto _error;
  }

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* SKIP - tidy up HDF ids */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export elements of a read alignment set hitting a reference sequence region.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Read alignment set to be exported\n"
  "  [--out-fmt|-o gff|sam|wig|bowtie|bowtie_extended|read_count]\n"
  "      Output file format; read_count prints a count of the matching reads\n"
  "  [--ref-id|-r 0]\n"
  "      Numeric identifier of reference sequence that exported reads aligned to \n"
  "      in the named alignment operation\n"
  "  [--ref-beg|-b 1]\n"
  "      Beginning of range on reference sequence (1-based inclusive)\n"
  "  [--ref-end|-e 10000]\n"
  "      End of range on reference sequence (1-based inclusive)\n"
  "  [--strand|-s +|-]\n"
  "      Strand on reference sequence; reads aligning to either strand are \n"
  "      exported if strand is not specified\n"
  "  [--window|-w <width>]\n"
  "      For wig-format output, apply windowing of the specified width \n"
  "  [--window-method|-m max|maximum|avg|average]\n"
  "      For windowed wig-format output, use either the maximum read depth or an \n"
  "      average across the window \n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_alignment -f bvt.h5 -a genome -r 17 -b 1000 -e 2000 -o wig -w 25 \\\n"
  "      -m max > max_depth.wig\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *window, char *window_method, int *is_all,
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:o:a:r:b:e:s:w:m:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument, is_verbose, 1},
    {"debug",    no_argument, is_debug, 1},
    {"all",      no_argument, is_all, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"out-fmt",  required_argument, 0, 'o'},
    {"aln",      required_argument, 0, 'a'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"strand",   required_argument, 0, 's'},
    {"window",   required_argument, 0, 'w'},
    {"window-method", required_argument, 0, 'm'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* output format */
    case 'o':
      H5UTL_CHECK_OPTARG;
      strcpy(out_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_id) != 1) goto USAGE;
      break;
      
      /* reference sequence range */
    case 'b':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_beg) != 1) goto USAGE;
      (*ref_beg)--;
      break;
      
      /* reference sequence range */
    case 'e':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_end) != 1) goto USAGE;
      (*ref_end)--;
      break;
      
      /* strand */
    case 's':
      if (optarg == NULL) goto USAGE;
      if (optarg[0] == '+') {
	*strand = 1;
      }
      else if (optarg[0] == '-') {
	*strand = -1;
      }
      else {
	*strand = 0;
      }
      break;
      
      /* window size for -o wig output format */
    case 'w':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", window) != 1) goto USAGE;
      break;
      
      /* aggregation method for window size > 1 */
    case 'm':
      H5UTL_CHECK_OPTARG;
      strcpy(window_method, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  if (strcmp(out_fmt, "sam") == 0 && *is_all) {
    fprintf(stderr, "'h5_export_alignment ... -o sam --all' is not currently supported; exiting...\n");
    exit(0);
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}



