/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *is_strict, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/


int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char out_fmt[40];
  char aln_name[40];
  int ref_id, ref_beg, ref_end, strand;
  int is_strict  = 0;
  int is_verbose = 0;
  int is_debug   = 0;


  hid_t fid = -1;
  hid_t gid = -1;
  hid_t level_did = -1;
  hid_t did_refmap = -1;
  hid_t did_annot = -1;

  /* Option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "annotation");
  strcpy(out_fmt,      "gff");
  ref_id  = 0;
  ref_beg = 1;
  ref_end = 300000000;
  strand  = 0;

  /* Parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, out_fmt, aln_name,
			&ref_id, &ref_beg, &ref_end, &strand, &is_strict, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* read from the HDF file */
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDONLY, H5P_DEFAULT));

  /* open the requested alignment group */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", ALIGNMENTS_GROUP, aln_name);
  HDFERR(gid = H5Gopen(fid, hdf_path, H5P_DEFAULT));

  /* open the refmap dataset */
  HDFERR(did_refmap = H5Dopen(gid, REFMAP_DATASET, H5P_DEFAULT));
  HDFERR(did_annot = H5Dopen(gid, ANNOTATION_DATASET, H5P_DEFAULT));
  
  /* open the NCList level dataset */
  level_did = H5Dopen(gid, NCLIST_LEVEL_DATASET, H5P_DEFAULT);

  /* search for reads that align to specified region of a ref_id */
  if (strcmp(out_fmt, "gff") == 0) {
    if (is_strict) {
      printf("##gff-version 3\n");
    }
    H5BIOrefmap_nclist_search(did_refmap, level_did, aln_name, 
			 strand, ref_id, ref_beg, ref_end, -1 /* seq_len */,
			 refmap_to_annot_gff, &did_annot);
  }

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export reference annotations GFF/GTF format in a specified region.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Name for this annotation track\n"
  "  [--out-fmt|-o gff]\n"
  "      Output file format for annotations\n"
  "  [--ref-id|-r 0]\n"
  "      Numeric identifier of reference sequence\n"
  "  [--ref-beg|-b 1]\n"
  "      Beginning of range on reference sequence (1-based inclusive)\n"
  "  [--ref-end|-e 10000]\n"
  "      End of range on reference sequence (1-based inclusive)\n"
  "  [--strand|-s +|-]\n"
  "      Strand on reference sequence; annotations on both strands are \n"
  "      exported if strand is not specified\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_annotation -f bvt.h5 -a refseq_genes -r 18 -b 2000 -e 4000 > refseq_genes_chr19_2001-4000.gtf\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, char *aln_name,
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *is_strict, 
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:o:a:r:b:e:s:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument, is_verbose, 1},
    {"debug",    no_argument, is_debug, 1},
    {"strict",   no_argument, is_strict, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"out-fmt",  required_argument, 0, 'o'},
    {"aln",      required_argument, 0, 'a'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"strand",   required_argument, 0, 's'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* output format */
    case 'o':
      H5UTL_CHECK_OPTARG;
      strcpy(out_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_id) != 1) goto USAGE;
      break;
      
      /* reference sequence range */
    case 'b':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_beg) != 1) goto USAGE;
      (*ref_beg)--;
      break;
      
      /* reference sequence range */
    case 'e':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_end) != 1) goto USAGE;
      (*ref_end)--;
      break;
      
      /* strand */
    case 's':
      if (optarg == NULL) goto USAGE;
      if (optarg[0] == '+') {
	*strand = 1;
      }
      else if (optarg[0] == '-') {
	*strand = -1;
      }
      else {
	*strand = 0;
      }
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}



