/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char *group_name, char *dataset_name, 
			  int *is_verbose, int *is_debug);

herr_t attr_info(hid_t loc_id, const char *name, const H5A_info_t *ainfo, void *op_data);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char group_name[HDF_OBJNAME_MAX];
  char dataset_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(group_name,   "/");
  strcpy(dataset_name, "dataset");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, group_name, dataset_name, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid, gid, did;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDONLY, H5P_DEFAULT));
  HDFERR(gid = H5Gopen(fid, group_name, H5P_DEFAULT));
  HDFERR(did = H5Dopen(gid, dataset_name, H5P_DEFAULT));

  /* iterate over all attributes, printing them */
  H5Aiterate(did, H5_INDEX_NAME, H5_ITER_NATIVE, NULL, attr_info, NULL);


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* terminating process, so don't worry about tidying up */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export all attributes from a dataset as tab-separated pairs on stdout.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      HDF5 file name\n"
  "  [--group|-g <group_path>]\n"
  "      HDF5 group path\n"
  "  [--dataset|-d <dataset>]\n"
  "      HDF5 dataset at that group path\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_attribute -f bvt.h5 -g /alignments/genome -d alignment > genome-attrs.txt\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char *group_name, char *dataset_name, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:g:d:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"group",    required_argument, 0, 'g'},
    {"dataset",  required_argument, 0, 'd'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* group name */
    case 'g':
      H5UTL_CHECK_OPTARG;
      strcpy(group_name, optarg);
      break;
      
      /* dataset name */
    case 'd':
      H5UTL_CHECK_OPTARG;
      strcpy(dataset_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


/* H5Aiterate operator */
herr_t attr_info(hid_t loc_id, const char *name, const H5A_info_t *ainfo, void *op_data) {

  hid_t attr, atype, aspace; /* attribute, datatype, dataspace identifiers */
  int rank;
  int i;
  hsize_t sdim[64]; 
  size_t npoints; /* number of elements in the array attribute */ 
  int point_out;    
  float *float_array; /* pointer to the array attribute */
  H5S_class_t class;

  /* open the attribute using its name */
  attr = H5Aopen_name(loc_id, name);

  /* print attribute name */
  printf("%s\t", name);

  /* get attribute datatype, dataspace, rank, and dimensions */
  atype = H5Aget_type(attr);
  aspace = H5Aget_space(attr);
  rank = H5Sget_simple_extent_ndims(aspace);
  H5Sget_simple_extent_dims(aspace, sdim, NULL);

  /* get dataspace type */
  class = H5Sget_simple_extent_type(aspace);

  if (H5T_INTEGER == H5Tget_class(atype)) {
    H5Aread(attr, atype, &point_out);
    printf("%d\n", point_out);
  }

  if (H5T_FLOAT == H5Tget_class(atype)) {
    npoints = H5Sget_simple_extent_npoints(aspace);
    float_array = (float *)malloc(sizeof(float) * (int)npoints); 
    H5Aread(attr, atype, float_array);
    for (i = 0; i < (int)npoints; i++) printf("%f ", float_array[i]); 
    printf("\n");
    free(float_array);
  }


  H5Tclose(atype);
  H5Sclose(aspace);
  H5Aclose(attr);

  return 0;
}
