/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char aln_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, aln_name, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDONLY, H5P_DEFAULT));

  /* open the alignments group */
  hid_t gid_alns, gid_aln, did_cl, sid_cl;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));
  HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));
  HDFERR(did_cl = H5Dopen(gid_aln, CLUSTER_DATASET, H5P_DEFAULT));
  HDFERR(sid_cl = H5Dget_space(did_cl));
  int cl_num = H5UTLget_1D_extent(sid_cl);

  /* use fixed-size buffer to iterate over the cluster dataset */
  int buf_size = HDF_RW_BUF_SIZE;
  cluster_t *cl_buf = (cluster_t *)calloc(buf_size, sizeof(cluster_t));

  /* create the cluster memory datatype */
  hid_t tid_cl;
  HDFERR(tid_cl = H5BIOcluster_type());

  /* iterate over all hits */
  int i;
  cluster_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < cl_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = cl_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_cl, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_cl, tid_cl, mid, sid_cl, H5P_DEFAULT, cl_buf);
      H5Sclose(mid);
    }
    item = cl_buf + i - buf_offset;

    /* export the cluster */
    printf("%d\t%d\t%d\t%d\n", 
	   item->ref_id,
	   item->beg_pos,
	   item->end_pos,
	   item->num_read);
  }


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export alignment cluster positions, annotating with reference info.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Read alignment set that has been clustered\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_cluster -f bvt.h5 -a genome > genome-cluster.txt\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"aln",      required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


