/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, 
			  char **filter_aln_names, int *num_filter, 
                          int **read_id, int *num_read_id, 
			  char *aln_name, int *is_seek, 
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char out_fmt[40];
  char *filter_aln_names[10]; /* TODO - should be unlimited */
  int num_filter;
  int *read_id = NULL;
  int num_read_id = 0;
  char aln_name[40];
  int ref_id, ref_beg, ref_end, strand;
  int is_fastq = 0;
  int is_seek = 0;
  int is_verbose = 0;
  int is_debug   = 0;
  hid_t level_did = -1;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  num_filter = 0;
  *aln_name = 0;
  ref_id  = 0;
  ref_beg = 1;
  ref_end = 10000;
  strand  = 0;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, out_fmt, 
			filter_aln_names, &num_filter, 
			&read_id, &num_read_id, 
			aln_name, &is_seek, 
			&ref_id, &ref_beg, &ref_end, &strand, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* boolean for fastq output format */
  if (strcmp(out_fmt, "fastq") == 0) {
    is_fastq = 1;
  }

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  /* read/write needed for possible creation of alignment readmask datasets */
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be available */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* open the alignments group */
  hid_t gid_alns;
  if (num_filter > 0 || 
      *aln_name != 0) {
    HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));
  }
  
  /* open the NCList level dataset */
  level_did = H5Dopen(gid_alns, NCLIST_LEVEL_DATASET, H5P_DEFAULT);

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_num, seq_len;
  H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);
  hid_t did_qual, sid_qual;
  if (is_fastq) {
    sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, QUALITY_DATASET);
    if ((did_qual = H5Dopen(fid, hdf_path, H5P_DEFAULT)) < 0) {
      fprintf(stderr, "Quality values not available for fastq output\n");
      goto _error;
    }
    HDFERR(sid_qual = H5Dget_space(did_qual));
  }

  /* form the OR-ed read-mask */
  char *full_readmask = (char *)calloc(seq_num, sizeof(char));
  char *one_readmask = (char *)calloc(seq_num, sizeof(char));

  /* optionally add a readmask from explicit read_id list passed */
  if (num_read_id > 0) {
    int i, j;
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] = 1;
    }
    for (i = 0; i < num_read_id; i++) {
      if (0 <= read_id[i] && read_id[i] < seq_num) {
	full_readmask[read_id[i]] = 0;
      }
    }
  }

  /* optionally add a readmask from alignment region lookup */
  if (*aln_name != 0) {
    int j;
    hid_t gid_aln, did_aln;
    HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));
    HDFERR(did_aln = H5Dopen(gid_aln, REFMAP_DATASET, H5P_DEFAULT));
    for (j = 0; j < seq_num; j++) {
      one_readmask[j] = 1;
    }
    H5BIOrefmap_nclist_search(did_aln, level_did, aln_name, 
			 strand, ref_id, ref_beg, ref_end, seq_len,
			 refmap_to_readmask, one_readmask);
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] |= one_readmask[j];
    }
  }

  /* add readmasks for each alignment dataset to be excluded */
  int i, j;
  hid_t gid_aln, did_mask;
  for (i = 0; i < num_filter; i++) {
    HDFERR(gid_aln = H5Gopen(gid_alns, filter_aln_names[i], H5P_DEFAULT));
    GENERR(H5BIOreadmask_alignment(fid, gid_aln));
    HDFERR(did_mask = H5Dopen(gid_aln, READMASK_DATASET, H5P_DEFAULT));
    H5Dread(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, one_readmask);
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] |= one_readmask[j];
    }
    H5Dclose(did_mask);
    H5Gclose(gid_aln);
  }

  if (is_seek) {
    /* seek to specified sequences */
    char *seq = (char *)calloc(seq_len + 1, sizeof(char));
    char *qual = (char *)calloc(seq_len + 1, sizeof(char));
    hsize_t slab_count[2] = { 1, seq_len };
    hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
    for (i = 0; i < seq_num; i++) {
      if (full_readmask[i] == 0) {
	/* seek to single sequence */
	hsize_t slab_start[2] = { i, 0 };
	H5Sselect_hyperslab(sid_seq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
	H5Dread(did_seq, H5T_NATIVE_CHAR, mid, sid_seq, H5P_DEFAULT, seq);
	*(seq + seq_len) = 0;

	if (is_fastq) { /* fastq format */
	  H5Sselect_hyperslab(sid_qual, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
	  H5Dread(did_qual, H5T_NATIVE_CHAR, mid, sid_qual, H5P_DEFAULT, qual);
	  *(qual + seq_len) = 0;
	  printf("@%d\n%s\n+%d\n%s\n", i, seq, i, qual);
	}
	else { /* fasta format */
	  printf(">%d\n%s\n", i, seq);
	}
      }
    }
    H5Sclose(mid);
  }
  else {
    /* filter the reads using the composite read-mask */
    int buf_size = HDF_RW_BUF_SIZE;
    char *seq_buf = (char *)calloc(buf_size * seq_len, sizeof(char));
    char *qual_buf = (char *)calloc(buf_size * seq_len, sizeof(char));

    /* scan over all sequences */
    char *seq = (char *)calloc(seq_len + 1, sizeof(char));
    char *qual = (char *)calloc(seq_len + 1, sizeof(char));
    int buf_offset = - buf_size;
    for (i = 0; i < seq_num; i++) {
      if (i >= buf_offset + buf_size) {
	/* fill buffer */
	buf_offset += buf_size;
	int buf_fill = seq_num - buf_offset;
	if (buf_fill > buf_size) {
	  buf_fill = buf_size;
	}
	hsize_t slab_start[2] = { buf_offset, 0 };
	hsize_t slab_count[2] = { buf_fill, seq_len };
	hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
	H5Sselect_hyperslab(sid_seq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
	H5Dread(did_seq, H5T_NATIVE_CHAR, mid, sid_seq, H5P_DEFAULT, seq_buf);
	if (is_fastq) {
	  H5Sselect_hyperslab(sid_qual, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
	  H5Dread(did_qual, H5T_NATIVE_CHAR, mid, sid_qual, H5P_DEFAULT, qual_buf);
	}
	H5Sclose(mid);
      }
      memcpy(seq, seq_buf + (i - buf_offset) * seq_len, seq_len);
      *(seq + seq_len) = 0;
      if (is_fastq) {
	memcpy(qual, qual_buf + (i - buf_offset) * seq_len, seq_len);
	*(qual + seq_len) = 0;
      }
      
      if (full_readmask[i] == 0) {
	if (is_fastq) {
	  printf("@%d\n%s\n+%d\n%s\n", i, seq, i, qual);
	}
	else {
	  printf(">%d\n%s\n", i, seq);
	}
      }
    }
  }


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export read sequences and, if fastq output format is selected, ASCII-encoded \n"
  "read qualities. The BioHDF-assigned numeric identifier is used in export. \n"
  "\n"
  "The list of exported reads is specified using one or more methods. Several \n"
  "methods of specifying reads can be combined such that reads excluded by any \n"
  "one method are excluded from export.\n"
  "  (-x) reads are excluded if they align in any of the specified alignment \n"
  "      operations;\n"
  "  (-d) reads are included if explicitly listed by numeric identifier;\n"
  "  (-a,-r,-b,-e,-s) reads are included if they align to the specified region and \n"
  "      strand in this alignment operation.\n"
  "\n"
  "The parameter --use-seek seeks for individual reads instead of scanning and \n"
  "filtering the whole set, and so improves performance when a small subset of \n"
  "reads are being exported. If the majority of reads are being exported, the \n"
  "default (full scan) method is more efficient.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name'\n"
  "  [--out-fmt|-i fasta|fastq]\n"
  "      Output file format for read sequences and associated ASCII-encoded \n"
  "      qualities (fastq format only); default 'fasta'\n"
  "  [--exclude|-x <alignment>]+\n"
  "      Names of one or more alignment operations available in this BioHDF file\n"
  "  [--read_id|-d <read_id>]+\n"
  "      Numeric identifiers of reads to be exported\n"
  "  [--aln|-a <alignment>]\n"
  "      Export reads that aligned in the named alignment operation\n"
  "  [--ref-id|-r 0]\n"
  "      Numeric identifier of reference sequence that exported reads aligned to\n"
  "      in the named alignment operation\n"
  "  [--ref-beg|-b 1]\n"
  "      Beginning of range on reference sequence (1-based inclusive)\n"
  "  [--ref-end|-e 10000]\n"
  "      End of range on reference sequence (1-based inclusive)\n"
  "  [--strand|-s +|-]\n"
  "      Strand on reference sequence; reads aligning to either strand are \n"
  "      exported if strand is not specified\n"
  "  [--use-seek]\n"
  "      Use seek method instead of default scan method; suitable for exporting \n"
  "      a minority of reads\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_read -f bvt.h5 -x junk > output-reads.fq\n"
  "  h5_export_read -f bvt.h5 -a genome -r 18 -b 123 -e 456 -s + > output-reads.fq\n"
  "  h5_export_read -f bvt.h5 -d 5 -d 17 -d 24 --use-seek > output-reads.fq\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *out_fmt, 
			  char **filter_aln_names, int *num_filter, 
			  int **read_id, int *num_read_id, 
			  char *aln_name, int *is_seek,
			  int *ref_id, int *ref_beg, int *ref_end, int *strand,
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:o:x:d:a:r:b:e:s:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"use-seek", no_argument,       is_seek, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"out-fmt",  required_argument, 0, 'o'},
    {"exclude",  required_argument, 0, 'x'},
    {"read_id",  required_argument, 0, 'd'},
    {"aln",      required_argument, 0, 'a'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"strand",   required_argument, 0, 's'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  int size_read_id = 0;

  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* output format */
    case 'o':
      H5UTL_CHECK_OPTARG;
      strcpy(out_fmt, optarg);
      break;
      
      /* alignment name for masking of reads */
    case 'x':
      H5UTL_CHECK_OPTARG;
      filter_aln_names[*num_filter] = (char *)calloc(HDF_OBJNAME_MAX, sizeof(char));
      strcpy(filter_aln_names[*num_filter], optarg);
      (*num_filter)++;
      break;
      
      /* read_id list - used as a filter if it exists */
    case 'd':
      H5UTL_CHECK_OPTARG;
      {
	int a_read_id;
	if (sscanf(optarg, "%d", &a_read_id) != 1) goto USAGE;
	if (*read_id == NULL) {
	  size_read_id = 1;
	  *read_id = (int *)malloc(size_read_id * sizeof(int));
	}
	else if (*num_read_id == size_read_id) {
	  size_read_id *= 2;
	  *read_id = (int *)realloc(*read_id, size_read_id * sizeof(int));
	}
	(*read_id)[*num_read_id] = a_read_id;
	(*num_read_id)++;
      }
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_id) != 1) goto USAGE;
      break;
      
      /* reference sequence range */
    case 'b':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_beg) != 1) goto USAGE;
      (*ref_beg)--;
      break;
      
      /* reference sequence range */
    case 'e':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_end) != 1) goto USAGE;
      (*ref_end)--;
      break;
      
      /* strand */
    case 's':
      if (optarg == NULL) goto USAGE;
      if (optarg[0] == '+') {
	*strand = 1;
      }
      else if (optarg[0] == '-') {
	*strand = -1;
      }
      else {
	*strand = 0;
      }
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


