/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char ref_name[HDF_OBJNAME_MAX];
  int ref_id, ref_beg, ref_end;
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(ref_name, "refdb");
  ref_id  = 0;
  ref_beg = 1;
  ref_end = 300000000;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, ref_name, 
			&ref_id, &ref_beg, &ref_end, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDONLY, H5P_DEFAULT));

  /* references must be available */
  if (!H5UTLgroup_exists(fid, REFERENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist; use 'h5_import_reference'\n", REFERENCES_GROUP);
    goto _error;
  }

  /* open the reference group */
  hid_t gid_refs, gid_ref;
  HDFERR(gid_refs = H5Gopen(fid, REFERENCES_GROUP, H5P_DEFAULT));
  HDFERR(gid_ref = H5Gopen(gid_refs, ref_name, H5P_DEFAULT));

  /* read sequence index dataset to memory */
  hid_t did_refidx, tid_refidx, sid_refidx;
  HDFERR(did_refidx = H5Dopen(gid_ref, REFIDX_DATASET, H5P_DEFAULT));
  HDFERR(sid_refidx = H5Dget_space(did_refidx));
  int refidx_num = H5UTLget_1D_extent(sid_refidx);
  refidx_t *refidx = (refidx_t *)calloc(refidx_num, sizeof(refidx_t));
  HDFERR(tid_refidx = H5BIOrefidx_type());
  H5Dread(did_refidx, tid_refidx, H5S_ALL, H5S_ALL, H5P_DEFAULT, refidx);
  
  long long refseq_offset = (*(refidx + ref_id)).offset;
  int refseq_length = (*(refidx + ref_id)).length;

  long long req_offset = refseq_offset + (long long)ref_beg;
  int req_length = ref_end - ref_beg + 1;

  if (req_length <= 0 || 
      req_offset + req_length > refseq_offset + refseq_length) {
    fprintf(stderr, "Bad parameters");
    exit(1);
  }

  /* select 1D char from dataset refseq in range req_offset to req_length */
  hid_t did_refseq, sid_refseq;
  HDFERR(did_refseq = H5Dopen(gid_ref, REFSEQ_DATASET, H5P_DEFAULT));
  HDFERR(sid_refseq = H5Dget_space(did_refseq));
  char *refseq = (char *)calloc(req_length + 1, sizeof(char));
  HDFRET(sid_refseq = H5Dget_space(did_refseq));
  hsize_t slab_start[1] = { req_offset };
  hsize_t slab_count[1] = { req_length };
  H5Sselect_hyperslab(sid_refseq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
  H5Dread(did_refseq, H5T_NATIVE_CHAR, mid, sid_refseq, H5P_DEFAULT, refseq);
  H5Sclose(mid);

  *(refseq + req_length) = '\0';

  printf("%s\n", refseq);


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Export specified region of a reference sequence from this database.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--ref|-d <reference>]\n"
  "      Name of reference database\n"
  "  [--ref-id|-r 0]\n"
  "      Numeric identifier of reference sequence\n"
  "  [--ref-beg|-b 1]\n"
  "      Beginning of range on reference sequence (1-based inclusive)\n"
  "  [--ref-end|-e 10000]\n"
  "      End of range on reference sequence (1-based inclusive)\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_export_reference -f bvt.h5 -d genome -r 18 -b 2000 -e 4000 > genome_chr19_2001-4000.fa\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *ref_id, int *ref_beg, int *ref_end, 
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:d:r:b:e:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"ref",      required_argument, 0, 'd'},
    {"ref-id",   required_argument, 0, 'r'},
    {"ref-beg",  required_argument, 0, 'b'},
    {"ref-end",  required_argument, 0, 'e'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* reference database */
    case 'd':
      H5UTL_CHECK_OPTARG;
      strcpy(ref_name, optarg);
      break;
      
      /* reference sequence internal id */
    case 'r':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_id) != 1) goto USAGE;
      break;
      
      /* reference sequence range */
    case 'b':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_beg) != 1) goto USAGE;
      (*ref_beg)--;
      break;
      
      /* reference sequence range */
    case 'e':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", ref_end) != 1) goto USAGE;
      (*ref_end)--;
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


