/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

void _write_indel_buf(hid_t did, int num, const char* buf, int buf_num, hid_t tid);

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char *aln_name, char *out_aln_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char aln_name[HDF_OBJNAME_MAX];
  char out_aln_name[HDF_OBJNAME_MAX];
  char *filter_aln_names[10]; /* TODO - should be unlimited */
  int num_filter;
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,  "alignment");
  strcpy(out_aln_name, "alignment_f");
  num_filter = 0;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, 
			aln_name, out_aln_name, 
			filter_aln_names, &num_filter, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be imported before alignment annotations added */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* open the alignments group */
  hid_t gid_alns;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int num_seq, seq_len;
  H5UTLget_2D_extent(sid_seq, &num_seq, &seq_len);
  H5Sclose(sid_seq);
  H5Dclose(did_seq);

  /* form the OR-ed read-mask */
  char *full_readmask = (char *)calloc(num_seq, sizeof(char));
  char *one_readmask = (char *)calloc(num_seq, sizeof(char));
  int i, j;
  hid_t gid_aln, did_mask;
  for (i = 0; i < num_filter; i++) {
    HDFERR(gid_aln = H5Gopen(gid_alns, filter_aln_names[i], H5P_DEFAULT));
    GENERR(H5BIOreadmask_alignment(fid, gid_aln));
    HDFERR(did_mask = H5Dopen(gid_aln, READMASK_DATASET, H5P_DEFAULT));
    H5Dread(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, one_readmask);
    for (j = 0; j < num_seq; j++) {
      full_readmask[j] |= one_readmask[j];
    }
    H5Dclose(did_mask);
    H5Gclose(gid_aln);
  }

  /* the alignment to be filtered */
  HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));

  /* Filter dataset 'indel', if it is present */
  int has_indel = 0;
  if (H5UTLdataset_exists(gid_aln, ALNINDEL_DATASET)) {
    has_indel = 1;
  }

  /* filter the alignment using the composite read-mask and, later, other 
     hit-based criteria */
  hid_t did_aln, did_indel;
  HDFERR(did_aln = H5Dopen(gid_aln, ALIGNMENT_DATASET, H5P_DEFAULT));
  if (has_indel) {
      HDFERR(did_indel = H5Dopen(gid_aln, ALNINDEL_DATASET, H5P_DEFAULT));
  }

  /* create the hit memory datatype */
  hid_t tid_hit;
  HDFERR(tid_hit = H5BIOhit_type());

  /* open named alignment group and dataset */
  if (H5UTLgroup_exists(gid_alns, out_aln_name)) {
    fprintf(stderr, "Group '/%s/%s' already exists, not overwriting\n", ALIGNMENTS_GROUP, out_aln_name);
    goto _error;
  }
  hid_t gid_out_aln, did_out_aln, did_out_indel;
  HDFERR(gid_out_aln = H5Gcreate(gid_alns, out_aln_name, H5P_DEFAULT, H5P_DEFAULT, 0));
  HDFERR(did_out_aln = H5UTLcreate_1D_dataset(gid_out_aln, ALIGNMENT_DATASET, tid_hit));
  if (has_indel) {
    HDFERR(did_out_indel = H5UTLcreate_2D_dataset(gid_out_aln, ALNINDEL_DATASET, H5T_NATIVE_CHAR, INDEL_MAX_LENGTH));
  }

  /* output buffer */
  int out_hit_num = 0;
  int out_buf_size = HDF_RW_BUF_SIZE;
  hit_t *out_hit_buf = (hit_t *)calloc(out_buf_size, sizeof(hit_t));
  char *out_indel_buf;
  if (has_indel) {
    out_indel_buf = (char *)calloc(out_buf_size * INDEL_MAX_LENGTH, sizeof(char));
  }

  /* iterate through the alignment dataset */
  hid_t sid_aln, sid_indel;
  HDFERR(sid_aln = H5Dget_space(did_aln));
  if (has_indel) {
    HDFERR(sid_indel = H5Dget_space(did_indel));
  }
  int hit_num = H5UTLget_1D_extent(sid_aln);

  /* use fixed-size buffer */
  int buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(buf_size, sizeof(hit_t));
  char *indel_buf;
  if (has_indel) {
    indel_buf = (char *)calloc(buf_size * INDEL_MAX_LENGTH, sizeof(char));
  }

  /* iterate over all hits */
  hit_t *item;
  int buf_offset = - buf_size;
  for (i = 0; i < hit_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = hit_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[1] = { buf_offset };
      hsize_t slab_count[1] = { buf_fill };
      H5Sselect_hyperslab(sid_aln, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dread(did_aln, tid_hit, mid, sid_aln, H5P_DEFAULT, hit_buf);
      H5Sclose(mid);

      if (has_indel) {
	hsize_t slab_start[2] = { buf_offset, 0 };
	hsize_t slab_count[2] = { buf_fill, INDEL_MAX_LENGTH };
	H5Sselect_hyperslab(sid_indel, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
	hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
	H5Dread(did_indel, H5T_NATIVE_CHAR, mid, sid_indel, H5P_DEFAULT, indel_buf);
	H5Sclose(mid);
      }
    }
    item = hit_buf + i - buf_offset;

    if (full_readmask[item->read_id] == 0) {

      /* TODO - apply any other per-hit filters */

      /* append the hit to dataset, via a buffer */
      int out_buf_loc = out_hit_num % out_buf_size;
      memcpy(out_hit_buf + out_buf_loc, item, sizeof(hit_t));
      ++out_hit_num;
      if (out_buf_loc == out_buf_size - 1) {
	write_hit_buf(did_out_aln, out_hit_num, out_hit_buf, out_buf_size, tid_hit);
      }

      if (has_indel) {
	memcpy(out_indel_buf + out_buf_loc * INDEL_MAX_LENGTH, 
	       indel_buf + (i - buf_offset) * INDEL_MAX_LENGTH, 
	       sizeof(char) * INDEL_MAX_LENGTH);
	if (out_buf_loc == out_buf_size - 1) {
	  _write_indel_buf(did_out_indel, out_hit_num, out_indel_buf, out_buf_size, H5T_NATIVE_CHAR);
	}
      }
    }
  }

  /* append last buffered hits to dataset */
  int out_buf_loc = out_hit_num % out_buf_size;
  if (out_buf_loc > 0) {
    write_hit_buf(did_out_aln, out_hit_num, out_hit_buf, out_buf_loc, tid_hit);

    if (has_indel) {
      _write_indel_buf(did_out_indel, out_hit_num, out_indel_buf, out_buf_loc, H5T_NATIVE_CHAR);
    }
  }

  /* store the 'num_read' statistic */
  int num_seq_filter = 0;
  for (j = 0; j < num_seq; j++) {
    if (full_readmask[j] == 0) ++num_seq_filter;
  }
  H5UTLwrite_attr_int(did_out_aln, "num_read", num_seq_filter);

  /* Create a hard link to the header dataset, if it is present */
  if (H5UTLdataset_exists(gid_aln, ALNHEADER_DATASET)) {
    H5Lcreate_hard(gid_aln, ALNHEADER_DATASET, gid_out_aln, ALNHEADER_DATASET, H5P_DEFAULT, H5P_DEFAULT);
  }  


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* RFTR - move to biohdf.c/h once h5_filter_alignment.c refactored */
void _write_indel_buf(hid_t did, int num_tot, const char* buf, int num_out, hid_t tid) {
  hsize_t dims[2] = { num_tot, INDEL_MAX_LENGTH };
  H5Dextend(did, dims);
  hid_t sid = H5Dget_space(did);
  
  hsize_t slab_start[2] = { num_tot - num_out, 0 };
  hsize_t slab_count[2] = { num_out, INDEL_MAX_LENGTH };
  H5Sselect_hyperslab(sid, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
  hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
  H5Dwrite(did, tid, mid, sid, H5P_DEFAULT, buf);
  H5Sclose(mid);
}


/* Command documentation */
const char *cmd_description = 
  "Filter a read alignment set by removing hits whose reads are present \n"
  "in one or more other alignment sets.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Read alignment set to be filtered\n"
  "  [--out-aln|-o <filtered_alignment>]\n"
  "      Name for filtered read alignment set\n"
  "  [--exclude|-x <mask_alignment>]\n"
  "      Read alignment sets that hit reads to be excluded during the filtering\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_filter_read -f bvt.h5 -a genome -x junk -o genome_minus_junk\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char *aln_name, char *out_aln_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:o:x:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"aln",      required_argument, 0, 'a'},
    {"out-aln",  required_argument, 0, 'o'},
    {"exclude",  required_argument, 0, 'x'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* filtered alignment name */
    case 'o':
      H5UTL_CHECK_OPTARG;
      strcpy(out_aln_name, optarg);
      if (optarg != NULL) strcpy(out_aln_name, optarg);
      break;
      
      /* alignment name for masking of reads */
    case 'x':
      H5UTL_CHECK_OPTARG;
      filter_aln_names[*num_filter] = (char *)calloc(HDF_OBJNAME_MAX, sizeof(char));
      strcpy(filter_aln_names[*num_filter], optarg);
      (*num_filter)++;
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


