/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
                          int *use_mapreads_id,
			  int *is_verbose, int *is_debug);

static int _parse_mapreads(hid_t did_aln, hid_t tid_hit, int *read_len, int use_mapreads_id);

static int _parse_bowtie(hid_t did_aln, hid_t tid_hit, int *read_len);

static int _parse_sam(hid_t gid_aln, hid_t did_aln, hid_t tid_hit, int *read_len);

static int _parse_bowtie_extended(hid_t did_aln, hid_t tid_hit, int *read_len);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char in_fmt[40];
  char aln_name[HDF_OBJNAME_MAX];
  int use_mapreads_id = 0;
  int is_verbose = 0;
  int is_debug   = 0;

  hid_t fid        = -1;
  hid_t gid_alns   = -1;
  hid_t gid_aln    = -1;
  hid_t did_aln    = -1;
  hid_t tid_hit    = -1;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(in_fmt,       "mapreads");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, in_fmt, aln_name, 
			&use_mapreads_id, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* append to the existing file using default properties */
  if (!file_exists(h5_file_name)) {
    HDFERR(fid = H5Fcreate(h5_file_name, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT));
  }
  else {
    HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));
  }

  /* create the alignments group, if not already present */
  HDFERR(gid_alns = H5UTLopen_or_create_group(fid, ALIGNMENTS_GROUP));

  /* open named alignment group */
  if (H5UTLgroup_exists(gid_alns, aln_name)) {
    fprintf(stderr, "Group '/%s/%s' already exists, not overwriting\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  HDFERR(gid_aln = H5Gcreate(gid_alns, aln_name, H5P_DEFAULT, H5P_DEFAULT, 0));

  /* create the hit memory datatype */
  HDFERR(tid_hit = H5BIOhit_type());

  /* create the "alignment" dataset (list of individual read/ref hits) */
  HDFERR(did_aln = H5UTLcreate_1D_dataset(gid_aln, ALIGNMENT_DATASET, tid_hit));

  /* get read length info if available */
  int *read_len;
  if (H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    /* open the "sequence" dataset for the total 'reads' statistic */
    hid_t gid_seqs, did_seq, sid_seq;
    HDFERR(gid_seqs = H5Gopen(fid, SEQUENCES_GROUP, H5P_DEFAULT));
    HDFERR(did_seq = H5Dopen(gid_seqs, SEQUENCE_DATASET, H5P_DEFAULT));
    HDFERR(sid_seq = H5Dget_space(did_seq));
    int seq_num, seq_len;
    H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);
    
    /* store the 'num_read' statistic */
    H5UTLwrite_attr_int(did_aln, "num_read", seq_num);

    /* get the read lengths */
    read_len = (int *)calloc(seq_num, sizeof(int));
    H5UTLget_2D_char_len(did_seq, read_len, seq_num);
  }


  /* For now, split the mapreads/csfasta parser from the SAM one. Lots of 
     common code can be abstracted out later. */
  if (strcmp(in_fmt, "mapreads") == 0) {
    GENERR(_parse_mapreads(did_aln, tid_hit, read_len, use_mapreads_id));
  }
  else if (strcmp(in_fmt, "bowtie") == 0) {
    GENERR(_parse_bowtie(did_aln, tid_hit, read_len));
  }
  else if (strcmp(in_fmt, "sam") == 0) {
    GENERR(_parse_sam(gid_aln, did_aln, tid_hit, read_len));
  }
  else if (strcmp(in_fmt, "bowtie_extended") == 0) {
    GENERR(_parse_bowtie_extended(did_aln, tid_hit, read_len));
  }

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* terminating process, so don't worry about cleaning up */
#if 0
  /* tidy up HDF ids */
  if (tid_hit    >= 0) H5Tclose(tid_hit);
  if (did_aln    >= 0) H5Dclose(did_aln);
  if (gid_aln    >= 0) H5Gclose(gid_aln); 
  if (gid_alns   >= 0) H5Gclose(gid_alns); 
  if (fid        >= 0) H5Fclose(fid); 
  /* free from the heap */
  free(hit_buf);
#endif

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Import the alignment of read sequences against a reference database. \n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Name for this read alignment operation\n"
  "  [--in-fmt|-i mapreads|sam|bowtie|bowtie_extended]\n"
  "      Input file format for read alignments\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_import_alignment -f bvt.h5 -a junk -i bowtie < bvt-reads-junk.concise \n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
                          int *use_mapreads_id,
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:i:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"in-fmt",   required_argument, 0, 'i'},
    {"aln",      required_argument, 0, 'a'},
    {"use-mapreads-id",    no_argument,       use_mapreads_id, 1},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* input format */
    case 'i':
      H5UTL_CHECK_OPTARG;
      strcpy(in_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


/* parse alignments in mapreads/csfasta format */
static int _parse_mapreads(hid_t did_aln, hid_t tid_hit, int *read_len, int use_mapreads_id) {

  char line[1024];
  char seqid_hits[400];
  int seq_num = 0;
  hit_buf_t hit_buf;
  hit_t hit;

  /* initialize the write buffer */
  H5BIOinit_hit_buf(&hit_buf, did_aln, tid_hit);

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip comments */
    if (line[0] == '#') {
      continue;
    }
    if (strncmp(line, "map start", 9) == 0) {
      continue;
    }

    /* parse line into read ID and ref_id hits */
    if (sscanf(line, ">%s", seqid_hits) == 1) {

      char *token = strtok(seqid_hits, ",");
      int read_id;
      if (use_mapreads_id) {
	if (sscanf(token, "%d", &read_id) != 1) {
	  fprintf(stderr, "File format problem:\n%s\n", line);
	  exit(1);
	}
      }	  
      while ((token = strtok(NULL, ","))) {

	int ref_id, position, mismatch;
	if (sscanf(token, "%d_%d.%d", &ref_id, &position, &mismatch) == 3) {
	} else if (sscanf(token, "%d.%d", &position, &mismatch) == 2) {
	  ref_id = 0;
	} else {
	  fprintf(stderr, "File format problem:\n%s\n", line);
	  exit(1);
	}
	
	/* ignores the read ID unless --use-mapreads-id set (assumes order is the same as the original csfasta file - big assumption!) */

	if (use_mapreads_id) {
	  hit.read_id  = read_id;
	} else {
	  hit.read_id  = seq_num;
	}
	hit.ref_id   = ref_id;
	hit.position = position;
	hit.length   = read_len[hit.read_id];
	hit.strand   = 1;
	if (hit.position < 0) {
	  hit.strand = -1;
	  hit.position = abs(hit.position);
	}
	hit.mismatch = mismatch;

        /* append the hit to dataset, via a buffer */
        H5BIOwrite_hit(&hit_buf, &hit);
      }
    }
    ++seq_num;
  }
  
  /* flush the write buffer */
  H5BIOflush_hit_buf(&hit_buf);

  return 0;
}


/* parse alignments in bowtie/concise format */
static int _parse_bowtie(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  hit_buf_t hit_buf;
  hit_t hit;

  /* initialize the write buffer */
  H5BIOinit_hit_buf(&hit_buf, did_aln, tid_hit);

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse line into read ID and ref_id hits */
    int read_id, ref_id, position, mismatch;
    char strand;
    if (sscanf(line, "%d%c:<%d,%d,%d>", 
	       &read_id, &strand, &ref_id, &position, &mismatch) == 5) {

      /* create the hit */
      hit.read_id  = read_id; /* 0-based index */
      hit.ref_id   = ref_id; /* 0-based index */
      hit.position = position; /* 0-based position */
      hit.length   = read_len[hit.read_id];
      hit.strand   = 1;
      if (strand == '-') {
	hit.strand = -1;
      }
      hit.mismatch = mismatch;
	
      /* append the hit to dataset, via a buffer */
      H5BIOwrite_hit(&hit_buf, &hit);
    }
  }
  
  /* flush the write buffer */
  H5BIOflush_hit_buf(&hit_buf);

  return 0;
}


/* parse alignments in an extended bowtie/concise format */
static int _parse_bowtie_extended(hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  hit_buf_t hit_buf;
  hit_t hit;

  /* initialize the write buffer */
  H5BIOinit_hit_buf(&hit_buf, did_aln, tid_hit);

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse line into read ID and ref_id hits */
    int read_id, ref_id, position, mismatch, length;
    char strand;
    if (sscanf(line, "%d%c:<%d,%d,%d,%d>", 
	       &read_id, &strand, &ref_id, &position, &mismatch, &length) == 6) {

      hit.read_id  = read_id; /* 0-based index */
      hit.ref_id   = ref_id; /* 0-based index */
      hit.position = position; /* 0-based position */
      hit.length   = length;
      hit.strand   = 1;
      if (strand == '-') {
	hit.strand = -1;
      }
      hit.mismatch = mismatch;
	
      /* append the hit to dataset, via a buffer */
      H5BIOwrite_hit(&hit_buf, &hit);
    }
  }
  
  /* flush the write buffer */
  H5BIOflush_hit_buf(&hit_buf);

  return 0;
}


/* parse alignments in SAM format */
static int _parse_sam(hid_t gid_aln, hid_t did_aln, hid_t tid_hit, int *read_len) {

  char line[1024];
  hit_buf_t hit_buf;
  hit_t hit;

  /* Create a string dataset for header information */
  int head_size = 0;
  hid_t did_head = H5UTLcreate_1D_dataset(gid_aln, ALNHEADER_DATASET, H5T_NATIVE_CHAR);

  /* Create a string array dataset for CIGAR indel information */
  int indel_num = 0;
  hid_t did_indel = H5UTLcreate_2D_dataset(gid_aln, ALNINDEL_DATASET, H5T_NATIVE_CHAR, INDEL_MAX_LENGTH);

  /* initialize the write buffer */
  H5BIOinit_hit_buf(&hit_buf, did_aln, tid_hit);

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* append headers to their dataset */
    if (line[0] == '@') {
      int head_line_size = strlen(line);
      hsize_t dims[1] = { head_size + head_line_size };
      H5Dextend(did_head, dims);
      hsize_t slab_start[1] = { head_size };
      hsize_t slab_count[1] = { head_line_size };
      hid_t sid_head = H5Dget_space(did_head);
      H5Sselect_hyperslab(sid_head, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(1, slab_count, slab_count); 
      H5Dwrite(did_head, H5T_NATIVE_CHAR, mid, sid_head, H5P_DEFAULT, line);
      H5Sclose(mid);
      H5Sclose(sid_head);
      head_size += head_line_size;
      continue;
    }

    /* parse line into read ID and ref_id hits */
    int qname, flag, rname, pos, mapq, mpos, isize;
    char cigar[200];
    char mrnm[80];
    char seq[1024];
    char qual[1024];
    
    if (sscanf(line, "%d\t%d\t%d\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s", &qname, &flag, &rname, &pos, &mapq, cigar, mrnm, &mpos, &isize, seq, qual) == 11) {

      if (flag & 0x4) continue;

      int mismatch = 0;
      char *nm_opt = strstr(line, "NM:i:");
      if (nm_opt != NULL) {
	if (sscanf(nm_opt, "NM:i:%d", &mismatch) != 1) {
	  mismatch = 0;
	}
      }

      hit.read_id  = qname;
      hit.ref_id   = rname;
      hit.position = pos - 1; /* to 0-based */
      hit.length   = read_len[hit.read_id]; /* TODO - parse this from cigar */
      hit.strand   = 1;
      if (flag & 0x10) hit.strand = -1;
      hit.mismatch = mismatch;
      if (mapq == 255) {
	hit.quality = -1;
      } else {
	hit.quality = mapq;
      }
	
      /* append the hit to dataset, via a buffer */
      H5BIOwrite_hit(&hit_buf, &hit);

      /* append CIGAR field to indel dataset */
      /* TODO - buffer this */
      int indel_length = strlen(cigar);
      if (indel_length >= INDEL_MAX_LENGTH) indel_length = INDEL_MAX_LENGTH - 1;
      hsize_t dims[2] = { indel_num + 1, INDEL_MAX_LENGTH };
      H5Dextend(did_indel, dims);
      hsize_t slab_start[2] = { indel_num, 0 };
      hsize_t slab_count[2] = { 1, indel_length };
      hid_t sid_indel = H5Dget_space(did_indel);
      H5Sselect_hyperslab(sid_indel, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
      H5Dwrite(did_indel, H5T_NATIVE_CHAR, mid, sid_indel, H5P_DEFAULT, cigar);
      H5Sclose(mid);
      H5Sclose(sid_indel);
      ++indel_num;
    }
  }
  
  /* flush the write buffer */
  H5BIOflush_hit_buf(&hit_buf);

  return 0;
}


