/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
			  int *is_verbose, int *is_debug);

static int _parse_gff(hid_t did_aln, hid_t tid_hit, hid_t did_annot);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char in_fmt[40];
  char aln_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;

  hid_t fid        = -1;
  hid_t gid_alns   = -1;
  hid_t gid_aln    = -1;
  hid_t did_aln    = -1;
  hid_t tid_hit    = -1;
  hid_t did_annot = -1;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(in_fmt,       "gff");
  strcpy(aln_name,     "annotation");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, in_fmt, aln_name, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open or append to the existing file using default properties */
  fid = H5UTLopen_or_create_file(h5_file_name);

  /* create the alignments group, if not already present */
  HDFERR(gid_alns = H5UTLopen_or_create_group(fid, ALIGNMENTS_GROUP));

  /* open named alignment group */
  if (H5UTLgroup_exists(gid_alns, aln_name)) {
    fprintf(stderr, "Group '/%s/%s' already exists, not overwriting\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  HDFERR(gid_aln = H5Gcreate(gid_alns, aln_name, H5P_DEFAULT, H5P_DEFAULT, 0));

  /* create the hit memory datatype */
  HDFERR(tid_hit = H5BIOhit_type());

  /* create the "alignment" dataset (used to store annotation intervals) */
  HDFERR(did_aln = H5UTLcreate_1D_dataset(gid_aln, ALIGNMENT_DATASET, tid_hit));

  /* create the "annotation" dataset (used to store annotation text) */
  HDFERR(did_annot = H5UTLcreate_2D_dataset(gid_aln, ANNOTATION_DATASET, H5T_NATIVE_CHAR, ANNOTATION_LENGTH));

  /* Parse the annotations */
  if (strcmp(in_fmt, "gff") == 0) {
    GENERR(_parse_gff(did_aln, tid_hit, did_annot));
  }
  else {
    fprintf(stderr, "Unknown input file format: %s\n", in_fmt);
    goto _error;
  }

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */
  if (tid_hit    >= 0) H5Tclose(tid_hit);
  if (did_aln    >= 0) H5Dclose(did_aln);
  if (gid_aln    >= 0) H5Gclose(gid_aln); 
  if (gid_alns   >= 0) H5Gclose(gid_alns); 
  if (fid        >= 0) H5Fclose(fid); 

  /* terminating process, so don't worry about freeing the heap */
  /* free(hit_buf); */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Import a reference annotation track in GFF/GTF format.\n"
  "The GFF/GTF column 1 (seqname) value must currently be numeric.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Name for this annotation track\n"
  "  [--in-fmt|-i gff]\n"
  "      Input file format for annotation track\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_import_annotation -f bvt.h5 -a refseq_genes < refseq_genes.gtf\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, char *aln_name, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:i:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"in-fmt",   required_argument, 0, 'i'},
    {"aln",      required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* input format */
    case 'i':
      H5UTL_CHECK_OPTARG;
      strcpy(in_fmt, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


/* parse alignments in GFF format */
#define MAX_GFF_ATTR_LEN 1024
static int _parse_gff(hid_t did_aln, hid_t tid_hit, hid_t did_annot) {

  /* create buffer for annotation data */
  int annot_buf_size = HDF_RW_BUF_SIZE;
  char* annot_buf = (char*)calloc(annot_buf_size * ANNOTATION_LENGTH, sizeof(char));

  char line[4096];
  int hit_num = 0;
  int hit_buf_size = HDF_RW_BUF_SIZE;
  hit_t *hit_buf = (hit_t *)calloc(hit_buf_size, sizeof(hit_t));
  hit_t hit;
  char *annot = (char *)calloc(ANNOTATION_LENGTH, sizeof(char));

  int feature_id = 0;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* skip headers */
    if (line[0] == '#') {
      continue;
    }

    /* parse line into read ID and ref_id hits */
    int ref_id, bloc1inc, eloc;
    char source[256];
    char feature[256];
    char score[80];
    char strand[4];
    char frame[4];
    char attr[MAX_GFF_ATTR_LEN];
    
    if (sscanf(line, "%d\t%s\t%s\t%d\t%d\t%s\t%s\t%s\t", 
	       &ref_id, source, feature, &bloc1inc, &eloc, score, strand, frame) == 8) {

      char *plasttab = strrchr(line, '\t');
      if (plasttab == NULL) return 1;
      strcpy(attr, plasttab + 1);
      char *pnewline = strchr(attr, '\n');
      if (pnewline != NULL) *pnewline = '\0';

      hit.read_id  = feature_id;
      hit.ref_id   = ref_id;
      hit.position = bloc1inc - 1; /* to 0-based */
      hit.length   = eloc - bloc1inc + 1;
      hit.strand   = 1;
      if (strand[0] == '-') hit.strand = -1;
      hit.mismatch = 0;
	
      /* append the hit to dataset, via a buffer */
      int buf_loc = hit_num % hit_buf_size;
      memcpy(hit_buf + buf_loc, &hit, sizeof(hit_t));
      ++hit_num;
      if (buf_loc == hit_buf_size - 1) {
	write_hit_buf(did_aln, hit_num, hit_buf, hit_buf_size, tid_hit);
      }

      /* store the feature and attr to annotation dataset via a buffer */
      memset(annot, 0, ANNOTATION_LENGTH);
      attr[MAX_GFF_ATTR_LEN - 1] = 0; /* nul-terminate for safety */

      /* integrate the feature and, optionally, the score */
      if (strcmp(score, ".") == 0) {
	sprintf(annot, "feature \"%s\"; %s", feature, attr);
      } 
      else {
	sprintf(annot, "feature \"%s\"; score \"%s\"; %s", feature, score, attr);
      }

      int annot_buf_loc = feature_id % annot_buf_size;
      memcpy(annot_buf + annot_buf_loc * ANNOTATION_LENGTH, annot, ANNOTATION_LENGTH);
      if (annot_buf_loc == annot_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_annot, feature_id + 1, annot_buf, annot_buf_size, ANNOTATION_LENGTH);
      }

      ++feature_id;
    }
  }
  
  /* append last buffered hits to dataset */
  int buf_loc = hit_num % hit_buf_size;
  if (buf_loc > 0) {
    write_hit_buf(did_aln, hit_num, hit_buf, buf_loc, tid_hit);
  }
  int annot_buf_loc = feature_id % annot_buf_size;
  H5UTLwrite_2D_char_buf(did_annot, feature_id, annot_buf, annot_buf_loc, ANNOTATION_LENGTH);

  return 0;
}

