/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, int *max_seq_length, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char in_fmt[40];
  int max_seq_length;
  int is_verbose = 0;
  int is_debug   = 0;


  hid_t fid;
  hid_t gid_seqs;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(in_fmt, "fasta");
  max_seq_length = -1; /* implies get from first sequence */

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, in_fmt, &max_seq_length,
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* create a new file using default properties */
  fid = H5UTLopen_or_create_file(h5_file_name);

  /* create the sequences group */
  if (H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '/%s' already exists, not overwriting\n", SEQUENCES_GROUP);
    goto _error;
  }
  HDFERR(gid_seqs = H5Gcreate(fid, SEQUENCES_GROUP, H5P_DEFAULT, H5P_DEFAULT, 0));
  H5Gset_comment(gid_seqs, ".", "Sequence collection");

  if (strcmp(in_fmt, "fasta") == 0 ||
      strcmp(in_fmt, "csfasta") == 0) {
    GENERR(H5BIOparse_fastA(gid_seqs, max_seq_length, is_verbose));
  } else if (strcmp(in_fmt, "fastq") == 0) {
    GENERR(H5BIOparse_fastQ(gid_seqs, max_seq_length, is_verbose));
  }
  

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Import read sequences provided in csfasta, fasta, or fastq format from stdin \n"
  "to a BioHDF file. ASCII-encoded base quality information is imported from \n"
  "fastq format.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name, defaults to 'file.h5'\n"
  "  [--in-fmt|-i csfasta|fasta|fastq]\n"
  "      Input file format for read sequences and associated ASCII-encoded \n"
  "      qualities (fastq format only); default 'fasta'\n"
  "  [--seq-len|-l <length>]\n"
  "      Maximum length of any read sequence to be stored; default is first \n"
  "      read defines the maximum length\n"
  "  [--verbose]\n"
  "      Display timing information about import process\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_import_read -f bvt.h5 -i fastq < input-reads.fq\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *in_fmt, int *max_seq_length,
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:i:l:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"in-fmt",   required_argument, 0, 'i'},
    {"seq-len",  required_argument, 0, 'l'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* input format */
    case 'i':
      H5UTL_CHECK_OPTARG;
      strcpy(in_fmt, optarg);
      break;
      
      /* max_seq_length (optional) */
    case 'l':
      H5UTL_CHECK_OPTARG;
      if (sscanf(optarg, "%d", max_seq_length) != 1) goto USAGE;
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


