/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *is_verbose, int *is_debug);

static int _parse_fasta(hid_t did_refid, hid_t did_refdef, hid_t tid_refidx, hid_t did_refidx, hid_t did_refseq, int is_verbose);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {
  
  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char ref_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(ref_name, "refdb");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, ref_name, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open h5 file */
  hid_t fid;
  if (file_exists(h5_file_name)) {
    HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));
  } else {
    HDFERR(fid = H5Fcreate(h5_file_name, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT));
  }

  /* create the references group */
  hid_t gid_refs, gid_ref;
  HDFERR(gid_refs = H5UTLopen_or_create_group(fid, REFERENCES_GROUP));
  HDFERR(gid_ref = H5Gcreate(gid_refs, ref_name, H5P_DEFAULT, H5P_DEFAULT, 0));

  /* create the datasets */
  hid_t did_refid, did_refdef, tid_refidx, did_refidx, did_refseq;
  HDFERR(did_refid = H5UTLcreate_2D_dataset(gid_ref, REFID_DATASET, H5T_NATIVE_CHAR, REFID_LENGTH));
  HDFERR(did_refdef = H5UTLcreate_2D_dataset(gid_ref, REFDEF_DATASET, H5T_NATIVE_CHAR, REFDEF_LENGTH));
  HDFERR(tid_refidx = H5BIOrefidx_type());
  HDFERR(did_refidx = H5UTLcreate_1D_dataset(gid_ref, REFIDX_DATASET, tid_refidx));
  HDFERR(did_refseq = H5UTLcreate_1D_dataset(gid_ref, REFSEQ_DATASET, H5T_NATIVE_CHAR));

  /* Parse the refdb fasta from STDIN */
  GENERR(_parse_fasta(did_refid, did_refdef, tid_refidx, did_refidx, did_refseq, is_verbose));
  

  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* terminating process, so don't worry about freeing the heap */
  
  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Import a reference sequence database from stdin to a BioHDF file.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name, defaults to 'file.h5'\n"
  "  [--ref|-d <reference>]\n"
  "      Name for this reference database\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_import_reference -f bvt.h5 -d genome < genome.fa\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:d:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"ref",      required_argument, 0, 'd'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* reference database */
    case 'd':
      H5UTL_CHECK_OPTARG;
      strcpy(ref_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


/* Parse fasta reference db to HDF. */
#define MAX_FASTA_LINE_LEN 1024
#define REFSEQ_INIT_BUF_SIZE 1024
static int _parse_fasta(hid_t did_refid, hid_t did_refdef, hid_t tid_refidx, hid_t did_refidx, hid_t did_refseq, int is_verbose) {

  /* create buffer for refid and refdef data */
  int refid_buf_size = HDF_RW_BUF_SIZE;
  char* refid_buf = (char*)calloc(refid_buf_size * REFID_LENGTH, sizeof(char));
  int refdef_buf_size = HDF_RW_BUF_SIZE;
  char* refdef_buf = (char*)calloc(refdef_buf_size * REFDEF_LENGTH, sizeof(char));

  /* parse the csfasta file */
  char line[MAX_FASTA_LINE_LEN]; /* TODO - parameterize max line length */
  char scanline[MAX_FASTA_LINE_LEN];
  char *refid = (char *)calloc(REFID_LENGTH, sizeof(char));
  char *refdef = (char *)calloc(REFDEF_LENGTH, sizeof(char));
  int ref_num = 0;

  long long refseq_offset = 0;
  int refseq_length = 0;
  int refseq_size = REFSEQ_INIT_BUF_SIZE;
  char *refseq = (char *)calloc(refseq_size, sizeof(char));

  refidx_t refidx_buf;

  while (fgetline(stdin, line, sizeof(line)) > 0) {

    /* parse id and def*/
    if (line[0] == '>') {
      /* write the previous sequence to datasets */
      if (refseq_length > 0) {
	H5UTLwrite_1D_char_buf(did_refseq, refseq, refseq_length);

	refidx_buf.offset = refseq_offset;
        refidx_buf.length = refseq_length;
	H5BIOwrite_refidx(did_refidx, ref_num, &refidx_buf, 1, tid_refidx);

	refseq_offset += (long long)refseq_length;
	refseq_length = 0;
	*refseq = '\0';
      }

      char *space;
      if ((space = strchr(line, 0x20)) != NULL) {
	memcpy(refid, line + 1, (int)(space - line - 1));
	*(refid + (int)(space - line - 1)) = 0;
	char *linebrk;
	if ((linebrk = strchr(line, 0x0A)) != NULL) *linebrk = 0;
	if ((linebrk = strchr(line, 0x0D)) != NULL) *linebrk = 0;
	*(line + MAX_FASTA_LINE_LEN - 1) = 0;
	strcpy(refdef, space + 1);
      } else {
	char *linebrk;
	if ((linebrk = strchr(line, 0x0A)) != NULL) *linebrk = 0;
	if ((linebrk = strchr(line, 0x0D)) != NULL) *linebrk = 0;
	*(line + MAX_FASTA_LINE_LEN - 1) = 0;
	strcpy(refid, line + 1);
	memset(refdef, 0, REFDEF_LENGTH);
      }
      
      /* append the refid to dataset, via a buffer */
      int refid_buf_loc = ref_num % refid_buf_size;
      memcpy(refid_buf + refid_buf_loc * REFID_LENGTH, refid, REFID_LENGTH);
      if (refid_buf_loc == refid_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_refid, ref_num + 1, refid_buf, refid_buf_size, REFID_LENGTH);
	memset(refid_buf, 0, refid_buf_size * REFID_LENGTH);
      }
      
      /* append the refdef to dataset, via a buffer */
      int refdef_buf_loc = ref_num % refdef_buf_size;
      memcpy(refdef_buf + refdef_buf_loc * REFDEF_LENGTH, refdef, REFDEF_LENGTH);
      if (refdef_buf_loc == refdef_buf_size - 1) {
	H5UTLwrite_2D_char_buf(did_refdef, ref_num + 1, refdef_buf, refdef_buf_size, REFDEF_LENGTH);
	memset(refdef_buf, 0, refdef_buf_size * REFDEF_LENGTH);
      }
      
      memset(refid, 0, REFID_LENGTH);
      memset(refdef, 0, REFDEF_LENGTH);
      
      ++ref_num;
    }
    /* parse sequence */
    else {
      if (sscanf(line, "%s", scanline)) {
	int scanline_len = strlen(scanline);
	if ((refseq_length + scanline_len + 1) > refseq_size) {
	  refseq_size *= 2;
	  refseq = (char *)realloc(refseq, refseq_size * sizeof(char));
	}
	strcpy(refseq + refseq_length, scanline);
	refseq_length += scanline_len;
      }
    }
  }
  /* write the previous sequence to datasets */
  if (refseq_length > 0) {
    H5UTLwrite_1D_char_buf(did_refseq, refseq, refseq_length);
    
    refidx_buf.offset = refseq_offset;
    refidx_buf.length = refseq_length;
    H5BIOwrite_refidx(did_refidx, ref_num, &refidx_buf, 1, tid_refidx);
  }
  
  /* flush the buffers */
  int refid_buf_loc = ref_num % refid_buf_size;
  H5UTLwrite_2D_char_buf(did_refid, ref_num, refid_buf, refid_buf_loc, REFID_LENGTH);
  int refdef_buf_loc = ref_num % refdef_buf_size;
  H5UTLwrite_2D_char_buf(did_refdef, ref_num, refdef_buf, refdef_buf_loc, REFDEF_LENGTH);

  return 0;
}

