/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Script description */
const char *description = 
  "Create an index dataset:\n"
  "    /alignments/<alignment>/refmap\n"
  "from the associated alignment dataset:\n"
  "    /alignments/<alignment>/alignment\n";


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
                          int *is_seq_qual, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char aln_name[HDF_OBJNAME_MAX];
  int is_seq_qual = 0;
  int is_verbose = 0;
  int is_debug   = 0;

  hid_t fid;
  hid_t gid_alns;
  hid_t gid_aln;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(aln_name,     "alignment");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, aln_name, &is_seq_qual, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open or create an HDF file using default properties */ 
  fid = H5UTLopen_or_create_file(h5_file_name);

  /* alignment group must exist */
  if (!H5UTLgroup_exists(fid, ALIGNMENTS_GROUP)) {
    fprintf(stderr, "Group '/%s/%s' must exist before indexing alignments; use 'h5_import_alignment'\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* open named alignment group */
  if (!H5UTLgroup_exists(gid_alns, aln_name)) {
    fprintf(stderr, "Group '/%s/%s' must exist before indexing alignments; use 'h5_import_alignment'\n", ALIGNMENTS_GROUP, aln_name);
    goto _error;
  }
  HDFERR(gid_aln = H5Gopen(gid_alns, aln_name, H5P_DEFAULT));

  /* create 'refmap' index dataset for this alignment */
  //GENERR(H5BIOindex_alignment(fid, gid_aln));
  GENERR(H5BIOindex_alignment_nclist(fid, gid_aln));


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* terminating, so don't worry about freeing the heap or closing HDF ids */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Index the alignment of read sequences against a reference database, \n"
  "to support rapid search and export. This index is generated automatically \n"
  "before export, if not already present, but this can take some time. \n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--aln|-a <alignment>]\n"
  "      Name for this read alignment operation\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_index_alignment -f bvt.h5 -a junk \n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *aln_name, 
                          int *is_seq_qual, 
                          int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"aln",      required_argument, 0, 'a'},
    {"sam",      no_argument,       is_seq_qual, 1},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name */
    case 'a':
      H5UTL_CHECK_OPTARG;
      strcpy(aln_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


