/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/


/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char *filter_aln_names[10]; /* TODO - should be unlimited */
  int num_filter;
  int is_verbose = 0;
  int is_debug   = 0;

  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  num_filter = 0;

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, 
			filter_aln_names, &num_filter, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDWR, H5P_DEFAULT));

  /* reads must be available */
  if (!H5UTLgroup_exists(fid, SEQUENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist before importing alignments; use 'h5_import_read'\n", SEQUENCES_GROUP);
    goto _error;
  }

  /* open the alignments group */
  hid_t gid_alns;
  HDFERR(gid_alns = H5Gopen(fid, ALIGNMENTS_GROUP, H5P_DEFAULT));

  /* get fixed sequence length from 2nd dim of 'sequence' dataset extent */
  /* TODO - refactor into biohdf.c */
  char hdf_path[PATH_MAX];
  sprintf(hdf_path, "/%s/%s", SEQUENCES_GROUP, SEQUENCE_DATASET);
  hid_t did_seq, sid_seq;
  HDFERR(did_seq = H5Dopen(fid, hdf_path, H5P_DEFAULT));
  HDFERR(sid_seq = H5Dget_space(did_seq));
  int seq_num, seq_len;
  H5UTLget_2D_extent(sid_seq, &seq_num, &seq_len);

  /* form the AND-ed read-mask */
  char *full_readmask = (char *)calloc(seq_num, sizeof(char));
  char *one_readmask = (char *)calloc(seq_num, sizeof(char));
  int i, j;
  hid_t gid_aln, did_mask;
  for (j = 0; j < seq_num; j++) {
    full_readmask[j] = 1;
  }
  for (i = 0; i < num_filter; i++) {
    HDFERR(gid_aln = H5Gopen(gid_alns, filter_aln_names[i], H5P_DEFAULT));
    GENERR(H5BIOreadmask_alignment(fid, gid_aln));
    HDFERR(did_mask = H5Dopen(gid_aln, READMASK_DATASET, H5P_DEFAULT));
    H5Dread(did_mask, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, one_readmask);
    for (j = 0; j < seq_num; j++) {
      full_readmask[j] &= one_readmask[j];
    }
    H5Dclose(did_mask);
    H5Gclose(gid_aln);
  }

  /* filter the reads using the composite read-mask */
  int buf_size = HDF_RW_BUF_SIZE;
  char *seq_buf = (char *)calloc(buf_size * seq_len, sizeof(char));

  /* iterate over all sequences */
  char *seq = (char *)calloc(seq_len + 1, sizeof(char));
  int buf_offset = - buf_size;
  for (i = 0; i < seq_num; i++) {
    if (i >= buf_offset + buf_size) {
      /* fill buffer */
      buf_offset += buf_size;
      int buf_fill = seq_num - buf_offset;
      if (buf_fill > buf_size) {
	buf_fill = buf_size;
      }
      hsize_t slab_start[2] = { buf_offset, 0 };
      hsize_t slab_count[2] = { buf_fill, seq_len };
      H5Sselect_hyperslab(sid_seq, H5S_SELECT_SET, slab_start, NULL, slab_count, NULL);
      hid_t mid = H5Screate_simple(2, slab_count, slab_count); 
      H5Dread(did_seq, H5T_NATIVE_CHAR, mid, sid_seq, H5P_DEFAULT, seq_buf);
      H5Sclose(mid);
    }
    memcpy(seq, seq_buf + (i - buf_offset) * seq_len, seq_len);
    *(seq + seq_len) = 0;

    if (full_readmask[i] == 1) {
      printf("%d\t%d\n", i, (int)(strlen(seq)));
    }
  }


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Provide information (read id, length, annotation) in tab-text format for reads \n"
  "that align in any of the listed alignment operations. \n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--include|-a <alignment>]+\n"
  "      Names of one or more alignment operations available in this BioHDF file\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_info_read -f bvt.h5 -a junk > bvt-reads-junk.info\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, 
			  char **filter_aln_names, int *num_filter, 
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:a:o:x:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"include",  required_argument, 0, 'a'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* alignment name for masking of reads */
    case 'a':
      H5UTL_CHECK_OPTARG;
      filter_aln_names[*num_filter] = (char *)calloc(HDF_OBJNAME_MAX, sizeof(char));
      strcpy(filter_aln_names[*num_filter], optarg);
      (*num_filter)++;
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


