/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "hdf5.h"
#include "hdfutil.h"
#include "biohdf.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/******************************************************************************
 * Local function declarations
 *****************************************************************************/

static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *is_verbose, int *is_debug);


/******************************************************************************
 * Exported function definitions
 *****************************************************************************/

int main(int argc, char** argv) {

  int return_code = 0;

  char h5_file_name[PATH_MAX];
  char ref_name[HDF_OBJNAME_MAX];
  int is_verbose = 0;
  int is_debug   = 0;


  /* option defaults */
  strcpy(h5_file_name, "file.h5");
  strcpy(ref_name, "refdb");

  /* parse command line options */
  GENERR(_parse_options(argc, argv, h5_file_name, ref_name, 
			&is_verbose, &is_debug) < 0);

  /* Switch off error handling unless requested */
  if (!is_debug) {
    H5Eset_auto2(H5E_DEFAULT, NULL, NULL);
  }

  /* open existing file using default properties */
  if (!file_exists(h5_file_name)) {
    fprintf(stderr, "File '%s' must exist\n", h5_file_name);
    goto _error;
  }
  hid_t fid;
  HDFERR(fid = H5Fopen(h5_file_name, H5F_ACC_RDONLY, H5P_DEFAULT));

  /* references must be available */
  if (!H5UTLgroup_exists(fid, REFERENCES_GROUP)) {
    fprintf(stderr, "Group '%s' must exist; use 'h5_import_reference'\n", REFERENCES_GROUP);
    goto _error;
  }

  /* open the reference group */
  hid_t gid_refs, gid_ref;
  HDFERR(gid_refs = H5Gopen(fid, REFERENCES_GROUP, H5P_DEFAULT));
  HDFERR(gid_ref = H5Gopen(gid_refs, ref_name, H5P_DEFAULT));

  /* read all info datasets to memory */
  hid_t did_refid, sid_refid;
  HDFERR(did_refid = H5Dopen(gid_ref, REFID_DATASET, H5P_DEFAULT));
  HDFERR(sid_refid = H5Dget_space(did_refid));
  int refid_num, refid_len;
  H5UTLget_2D_extent(sid_refid, &refid_num, &refid_len);
  char *refid = (char *)calloc(refid_num * refid_len, sizeof(char));
  H5Dread(did_refid, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, refid);

  hid_t did_refdef, sid_refdef;
  HDFERR(did_refdef = H5Dopen(gid_ref, REFDEF_DATASET, H5P_DEFAULT));
  HDFERR(sid_refdef = H5Dget_space(did_refdef));
  int refdef_num, refdef_len;
  H5UTLget_2D_extent(sid_refdef, &refdef_num, &refdef_len);
  char *refdef = (char *)calloc(refdef_num * refdef_len, sizeof(char));
  H5Dread(did_refdef, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, refdef);

  hid_t did_refidx, tid_refidx, sid_refidx;
  HDFERR(did_refidx = H5Dopen(gid_ref, REFIDX_DATASET, H5P_DEFAULT));
  HDFERR(sid_refidx = H5Dget_space(did_refidx));
  int refidx_num = H5UTLget_1D_extent(sid_refidx);
  refidx_t *refidx = (refidx_t *)calloc(refidx_num, sizeof(refidx_t));
  HDFERR(tid_refidx = H5BIOrefidx_type());
  H5Dread(did_refidx, tid_refidx, H5S_ALL, H5S_ALL, H5P_DEFAULT, refidx);

  int ref_id;
  for (ref_id = 0; ref_id < refidx_num; ref_id++) {
    printf("%d\t%d\t%s\t%s\n", 
	   ref_id, 
	   (*(refidx + ref_id)).length,
	   refid + ref_id * refid_len,
	   refdef + ref_id * refdef_len);
  }
  


  goto _no_error;
 _error:
  H5UTL_ERR_MSG(argv[0], is_debug);
  return_code = 1;

 _no_error:

  /* tidy up HDF ids */

  /* terminating process, so don't worry about freeing the heap */

  /* Indicate successful or failed run */
  return return_code;
}


/******************************************************************************
 * Local function definitions
 *****************************************************************************/

/* Command documentation */
const char *cmd_description = 
  "Provide reference database information in tab-text format:\n"
  "    BioHDF identifier, sequence length, original identifier, annotation.\n";

const char *cmd_option = 
  "  [--h5-file|-f <filename>]\n"
  "      BioHDF file name\n"
  "  [--ref|-d <reference>]\n"
  "      Name for this reference database\n"
  "  [--debug]\n"
  "      Display detailed error reporting\n";

const char *cmd_example = 
  "  h5_info_reference -f bvt.h5 -d genome > genome-info.txt\n";


/* Parse command line options */
static int _parse_options(int argc, char *argv[],
                          char *h5_file_name, char *ref_name, 
			  int *is_verbose, int *is_debug) {
  int c;
  int option_index = 0;

  /* Short options */
  char *options = "f:d:h";

  /* Long options */
  struct option long_options[] = {
    {"verbose",  no_argument,       is_verbose, 1},
    {"debug",    no_argument,       is_debug, 1},
    {"h5-file",  required_argument, 0, 'f'},
    {"ref",      required_argument, 0, 'd'},
    {"help",     no_argument,       0, 'h'},
    {0, 0, 0, 0}
  };
  
  /* Parse options */
  opterr = 0;
  for (;;) {      
    c = getopt_long(argc, argv, options, long_options, &option_index);
    if (-1 == c) {
      break;
    }
    switch (c) {
    case 0:
      /* If this option sets a flag, do nothing else now. */
      break;
      
      /* h5 file */
    case 'f':
      H5UTL_CHECK_OPTARG;
      strcpy(h5_file_name, optarg);
      break;
      
      /* reference database */
    case 'd':
      H5UTL_CHECK_OPTARG;
      strcpy(ref_name, optarg);
      break;
      
      /* Error, so show help and return error code */
    default : 
      fprintf(stderr, "\nError: Unrecognised command-line option.\n\n");
      /* fall-thru to help */
      
      /* help, show usage and exit */
    case 'h':
      goto USAGE;
    }
  }

  return 0;

 USAGE:
  fprintf(stderr, "\n%s\nOptions:\n%s\nExample:\n%s\n%s\n", 
	  cmd_description, cmd_option, cmd_example, CMD_LINK);
  exit(0);
}


