/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group and Geospiza, Inc.                             *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of BioHDF.  The full BioHDF copyright notice, including *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING.  COPYING can be found at the root of the source code   *
 * distribution tree.  If you do not have access to this file, you may       *
 * request a copy from help@hdfgroup.org.                                    *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/******************************************************************************
 *
 * HDF5 general-purpose utilities used in the BioHDF project. 
 * 
 * Some of these will be replaced with the H5LT (Lite) functionality.
 * 
 * Other functions, if generally useful, may be migrated into the main HDF
 * libraries by The HDF Group.
 *
 *****************************************************************************/

#ifndef HDFUTIL_H
#define HDFUTIL_H 1

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "hdf5.h"


/******************************************************************************
 * Definitions and types
 *****************************************************************************/

/* Maximum length of group or dataset name. */
#define HDF_OBJNAME_MAX 80

/* Default chunk size in slowest-changing dimension. */
#define HDF_CHUNK_SIZE 1024

/* Default deflate compression parameter for chunked datasets. */
#define HDF_COMPRESSION_LEVEL 5 /* from 0 to 9 */

/* Default size of read or write buffer. */
#define HDF_RW_BUF_SIZE 1024

/* Group name for file storage area. */
#define FILES_GROUP "files"

#define HDFERR(exp) if ((exp) < 0) { goto _error; }
#define HDFRET(exp) if ((exp) < 0) { return -1; }
#define GENERR(exp) if (exp) { goto _error; }

#define H5UTL_ERR_MSG(prog, is_debug) { fprintf(stderr, "%s: fatal error, exiting... \n", prog); if (!is_debug) fprintf(stderr, "  use --debug option for detailed error reporting\n"); }

#define H5UTL_CHECK_OPTARG if (optarg == NULL || optarg[0] == '-') goto USAGE


/******************************************************************************
 * Exported function declarations
 *****************************************************************************/

/* Read line from file pointer pf into s, returning length read */
int fgetline(FILE *pf, char s[], int lim);

/* Does this file exist? */
int file_exists (const char *file_name);

/* Does this HDF5 group exist? */
int H5UTLgroup_exists(hid_t loc_id, const char *name);

/* Does this HDF5 dataset exist? */
int H5UTLdataset_exists(hid_t loc_id, const char *name);

/* Open this HDF5 file, creating if it doesn't exist */
hid_t H5UTLopen_or_create_file(const char *file_name);

/* Open this group, creating if it doesn't exist */
hid_t H5UTLopen_or_create_group(hid_t loc_id, const char *name);

/* Create a one-dimension dataset of type. */
hid_t H5UTLcreate_1D_dataset(hid_t gid, const char *name, hid_t tid);

/* Create a one-dimension dataset of type with initial size. */
hid_t H5UTLcreate_1D_dataset_size(hid_t gid, const char *name, hid_t tid, int init_size);

/* Create a two-dimension dataset of type, with fastest-changing dimension len. */
hid_t H5UTLcreate_2D_dataset(hid_t gid, const char *name, hid_t tid, int len);

/* Create a two-dimension dataset of type, with fastest-changing dimension len with initial size of slower-changing dimension. */
hid_t H5UTLcreate_2D_dataset_size(hid_t gid, const char *name, hid_t tid, int len, int init_size);

/* Get size of one-dimension dataset. */
int H5UTLget_1D_extent(hid_t sid);

/* Get size of two-dimension dataset. */
void H5UTLget_2D_extent(hid_t sid, int *extent_0, int *extent_1);

/* Write from memory buffer to 1D dataset of type H5T_NATIVE_CHAR. */
void H5UTLwrite_1D_char_buf(hid_t did, const char* buf, int buf_num);

/* Read from 1D dataset of type H5T_NATIVE_CHAR to memory buffer. */
int H5UTLread_1D_char_buf(hid_t did, int pos, char* buf, int buf_num);

/* Write from memory buffer to 2D dataset of type H5T_NATIVE_CHAR. */
void H5UTLwrite_2D_char_buf(hid_t did, int num, const char* buf, int buf_num, int len);

/* Read string from a 2D character dataset. */
int H5UTLread_2D_char_line(hid_t did, int num, char* buf, int buf_len);

/* Get the lengths of all nul-terminated strings in this dataset. */
void H5UTLget_2D_char_len(hid_t did, int *len, int num);

/* TODO - possibly replace this with H5LTset_attribute_int */
/* Attach value to dataset as an integer attribute with given name. */
void H5UTLwrite_attr_int(hid_t did, const char *name, int value);

/* qsort data in buffer using comparator function and store to 1D dataset */
void H5UTLsort_store_1D_dataset(hid_t did, hid_t tid, void *buf, int num, int typelen, int(*comparator)(const void *, const void *));


#endif /* HDFUTIL_H */
